""" all styling of buttons and other general widgets, some defined colors... """
from typing import Callable, Optional, Any
from PySide6.QtWidgets import QPushButton, QLabel, QSizePolicy, QMessageBox, QLayout, QWidget, QMenu, \
                              QVBoxLayout, QHBoxLayout, QGridLayout, QFormLayout, QComboBox, QScrollArea # pylint: disable=no-name-in-module
from PySide6.QtGui import QImage, QPixmap, QAction, QKeySequence, QMouseEvent               # pylint: disable=no-name-in-module
from PySide6.QtCore import QByteArray, Qt           # pylint: disable=no-name-in-module
from PySide6.QtSvgWidgets import QSvgWidget         # pylint: disable=no-name-in-module
import qtawesome as qta
from qt_material import get_theme
from .backend import Backend
from .handleDictionaries import dict2ul

space = {'0':0, 's':5, 'm':10, 'l':20, 'xl':200} #spaces: padding and margin

def getColor(backend:Backend, color:str) -> str:
  """
  get color from theme
  - Python access: theme = get_theme(themeName)
  - For dark-blue:
     - {'primaryColor': '#448aff', 'primaryLightColor': '#83b9ff', 'secondaryColor': '#232629', 'secondaryLightColor': '#4f5b62',
     - 'secondaryDarkColor': '#31363b', 'primaryTextColor': '#000000', 'secondaryTextColor': '#ffffff'}

  Args:
    backend (Pasta): backend instance
    color (str): color to get [primary, primaryLight, secondary, secondaryLight, secondaryDark, primaryText, secondaryText]

  Returns:
    str: #123456 color code
  """
  if not hasattr(backend, 'configuration') or backend.configuration['GUI']['theme']=='none':
    if color=='primary':
      return '#000000'
    else:
      return '#BBBBBB'
  themeName = backend.configuration['GUI']['theme']
  return get_theme(f'{themeName}.xml')[f'{color}Color']


class TextButton(QPushButton):
  """ Button that has only text"""
  def __init__(self, label:str, widget:QWidget, command:list[Any]=[], layout:Optional[QLayout]=None,
               tooltip:str='', checkable:bool=False, style:str='', hide:bool=False, iconName:str=''):
    """
    Args:
      label (str): label printed on button
      widget (QWidget): widget / dialog that host the button and that has the execute function
      command (enum): command that is used in called-function: possibly a list of multiple terms
      layout (QLayout): button to be added to this layout
      tooltip (str): tooltip shown when mouse hovers the button
      checkable (bool): can the button change its background color
      style (str): css style
      hide (bool): hidden or shown initially
    """
    super().__init__()
    self.setText(label)
    self.setCheckable(checkable)
    self.setChecked(checkable)
    self.clicked.connect(lambda: widget.execute(command))
    if tooltip:
      self.setToolTip(tooltip)
    if style:
      self.setStyleSheet(style)
    else:
      primaryColor   = getColor(widget.comm.backend, 'primary')
      secondaryColor = getColor(widget.comm.backend, 'secondary')
      self.setStyleSheet(f'border-width: 0px; background-color: {primaryColor}; color: {secondaryColor}')
    if hide:
      self.hide()
    if iconName:
      color = 'black' if widget is None else getColor(widget.comm.backend, 'primary')
      icon = qta.icon(iconName, color=color, scale_factor=1)
      self.setIcon(icon)
    if layout is not None:
      layout.addWidget(self)


class IconButton(QPushButton):
  """ Button that has only an icon"""
  def __init__(self, iconName:str, widget:QWidget, command:list[Any]=[], layout:Optional[QLayout]=None,
               tooltip:str='', style:str='', hide:bool=False):
    """
    Args:
      iconName (str): icon to show on button
      widget (QWidget): widget / dialog that host the button and that has the execute function
      command (enum): command that is used in called-function: possibly a list of multiple terms
      layout (QLayout): button to be added to this layout
      tooltip (str): tooltip shown when mouse hovers the button
      style (str): css style
      hide (bool): hidden or shown initially
    """
    super().__init__()
    color = 'black' if widget is None else getColor(widget.comm.backend, 'primary')
    icon = qta.icon(iconName, color=color, scale_factor=1)  #color change here
    self.setIcon(icon)
    self.clicked.connect(lambda: widget.execute(command))
    self.setFixedHeight(30)
    if tooltip:
      self.setToolTip(tooltip)
    if style:
      self.setStyleSheet(style)
    else:
      self.setStyleSheet("border-width:0")
    if hide:
      self.hide()
    if layout is not None:
      layout.addWidget(self)


class Action(QAction):
  """ QAction and assign function to menu"""
  def __init__(self, label:str, widget:QWidget, command:list[Any],
               menu:QMenu, shortcut:Optional[str]=None, icon:str=''):
    """
    Args:
      label (str): label printed on submenu
      widget (QWidget): widget / dialog that host the button and that has the execute function
      command (enum): command that is used in called-function: possibly a list of multiple terms
      menu (QMenu): button to be added to this menu
      shortcut (str): shortcut (e.g. Ctrl+K)
      icon (str): icon name
    """
    super().__init__()
    self.setParent(widget)
    self.setText(label)
    self.triggered.connect(lambda : widget.execute(command))
    if icon:
      color = 'black' if widget is None else getColor(widget.comm.backend, 'secondaryText')
      self.setIcon(qta.icon(icon, color=color, scale_factor=1))
    if shortcut is not None:
      self.setShortcut(QKeySequence(shortcut))
    menu.addAction(self)


class Image():
  """ Image widget depending on type of data """
  def __init__(self, data:str, layout:Optional[QLayout], width:int=-1, height:int=-1, anyDimension:int=-1):
    """
    Args:
      data (str): image data in byte64-encoding or svg-encoding
      layout (QLayout): to be added to this layout
      width (int): width of image, dominant if both are given
      height (int): height of image
      anyDimension (int): maximum size in any direction
    """
    if data.startswith('data:image/'): #jpg or png image
      byteArr = QByteArray.fromBase64(bytearray(data[22:] if data[21]==',' else data[23:], encoding='utf-8'))
      imageW = QImage()
      imageType = data[11:15].upper()
      imageW.loadFromData(byteArr, format=imageType[:-1] if imageType.endswith(';') else imageType)   # type: ignore
      pixmap = QPixmap.fromImage(imageW)
      if height>0:
        pixmap = pixmap.scaledToHeight(height)
      if width>0:
        pixmap = pixmap.scaledToWidth(width)
      if anyDimension>0:
        if pixmap.size().height()>pixmap.size().width():
          pixmap = pixmap.scaledToHeight(anyDimension)
        else:
          pixmap = pixmap.scaledToWidth(anyDimension)
      label = QLabel()
      label.setPixmap(pixmap)
      label.setAlignment(Qt.AlignCenter) # type: ignore
      if layout is not None:
        layout.addWidget(label, alignment=Qt.AlignHCenter)  # type: ignore
    elif data.startswith('<?xml'): #svg image
      imageW = QSvgWidget()
      policy = imageW.sizePolicy()
      policy.setHorizontalPolicy(QSizePolicy.Fixed)
      policy.setVerticalPolicy(QSizePolicy.Fixed)
      imageW.setSizePolicy(policy)
      imageW.renderer().load(bytearray(data, encoding='utf-8'))
      if height>0:
        imageW.setMaximumSize(int(float(imageW.width())/float(imageW.height())*height) ,height)
      if width>0:
        imageW.setMaximumSize(width, int(float(imageW.height())/float(imageW.width())*width))
      if anyDimension>0:
        if imageW.height()>imageW.width():
          imageW.setMaximumSize(int(float(imageW.width())/float(imageW.height())*anyDimension) ,anyDimension)
        else:
          imageW.setMaximumSize(anyDimension, int(float(imageW.height())/float(imageW.width())*anyDimension))
      if layout is not None:
        layout.addWidget(imageW, alignment=Qt.AlignHCenter) # type: ignore
    elif len(data)>2:
      print(f'WidgetProjectLeaf:What is this image |{data[:50]}|')
    return


class Label(QLabel):
  """ Label widget: headline, ... """
  def __init__(self, text:str='', size:str='', layout:Optional[QLayout]=None,
               function:Optional[Callable[[str, str],None]]=None, docID:str='', tooltip:str=''):
    """
    Args:
      text (str): text on label
      size (str): size ['h1','h2','h3']
      layout (QLayout): layout to which to add the label
      function (function): function to call on mouse click
      docID (str): docID on other string to connect to this label
      tooltip (str): tooltip shown when mouse hovers the button
    """
    super().__init__()
    self.setText(text)
    if size == 'h1':
      self.setStyleSheet('font-size: 14pt')
    elif size == 'h2':
      self.setStyleSheet('font-size: 12pt')
    elif size == 'h3':
      self.setStyleSheet('font-size: 10pt')
    if layout is not None:
      layout.addWidget(self)
    self.mouseFunction = function
    self.identifier = docID
    if tooltip != '':
      self.setToolTip(tooltip)
    return

  def mousePressEvent(self, e:QMouseEvent) -> None:
    """
    Event after mouse press: only use internal members, not the event itself
    """
    if self.mouseFunction is not None:
      self.mouseFunction(self.text(), self.identifier)
    return



def showMessage(parent:QWidget, title:str, text:str, icon:str='', style:str='') -> None:
  """
  Show simple message box for little information

  Args:
    parent (QWidget): parent widget (self)
    title (str): title of box
    text (str): text in box
    icon (str): icon: 'Information','Warning','Critical'
    style (str): css style
  """
  dialog = QMessageBox(parent)
  dialog.setWindowTitle(title)
  dialog.setText(text)
  if not (text.startswith('<') and text.endswith('>')):
    dialog.setTextFormat(Qt.MarkdownText)
  if icon in {'Information', 'Warning', 'Critical'}:
    dialog.setIcon(getattr(QMessageBox, icon))
  if style!='':
    dialog.setStyleSheet(style)
  dialog.exec()
  return


class ScrollMessageBox(QMessageBox):
  """ Scrollable message box for lots of dictionary information """
  def __init__(self, title:str, content:dict[str,Any], style:str=''):
    """
    Args:
      title (str): title
      content (dict): dictionary of lots of information
      style (str): css style
    """
    cssStyle = '<style> ul {list-style-type: none; padding-left: 0; margin: 0; text-indent: -20px; padding-left: -20px;} </style>'
    QMessageBox.__init__(self)
    self.setWindowTitle(title)
    if style == '':
      self.setStyleSheet('QScrollArea{min-width:300 px; min-height: 400px}')
    else:
      self.setStyleSheet(style)
    scroll = QScrollArea(self)
    scroll.setWidgetResizable(True)
    self.content = QLabel()
    self.content.setWordWrap(True)
    self.content.setText(cssStyle+dict2ul(content))
    self.content.setTextInteractionFlags(Qt.TextSelectableByMouse)        # type: ignore[arg-type]
    scroll.setWidget(self.content)
    self.layout().addWidget(scroll, 0, 0, 1, self.layout().columnCount()) # type: ignore[call-arg]


def widgetAndLayout(direction:str='V', parentLayout:Optional[QLayout]=None, spacing:str='0', left:str='0',
                    top:str='0', right:str='0', bottom:str='0') -> tuple[QWidget, QLayout]:
  """
  Convenient function for widget and a boxLayout

  Spacings and margins:
  - different than in css/html
  - spacing is the space between elements in the orientation of the BoxLayout
  - is the padding that surrounds the content in the layout

  Distances are given in
  - '0': zero distance
  - 's': small distance used as padding round elements, or vertical spacings
  - 'm': medium used as space between horizontal elements
  - 'l': large used when things need to be separated
  - 'xl': extra large indentations, frames

  Args:
    direction (str): type of layout [H,V,Grid,Form]
    parentLayout (QLayout): to which layout should the widget be added. If none, no adding
    spacing (str): spacing
    left (str): padding on left
    top (str): padding on top
    right (str): padding on right
    bottom (str): padding on bottom
  """
  widget = QWidget()
  if direction=='V':
    layout = QVBoxLayout(widget)
  elif direction=='H':
    layout = QHBoxLayout(widget)
  elif direction=='Form':
    layout = QFormLayout(widget)
  else:
    layout = QGridLayout(widget)
  layout.setSpacing(space[spacing])
  layout.setContentsMargins(space[left], space[top], space[right], space[bottom])
  if parentLayout is not None:
    parentLayout.addWidget(widget)
  return widget, layout


def addRowList(layout:QLayout, label:str, default:str, itemList:list[str]) -> QComboBox:
  """
  Add a row with a combo-box to the form

  Args:
    layout (QLayout): layout to add row to
    label (str): label used in form
    default (str): default value
    itemList (list(str)): items to choose from

  Returns:
    QCombobox: filled combobox
  """
  widget = QComboBox()
  widget.addItems(itemList)
  widget.setCurrentText(default)
  layout.addRow(QLabel(label), widget)
  return widget



""" Terminology Lookup service """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: terminology_lookup_service.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import logging
from functools import reduce
from json import load
from os import getcwd
from os.path import dirname, join, realpath
from typing import Any

from pasta_eln.webclient.http_client import AsyncHttpClient


class TerminologyLookupService:
  """
  Terminology Lookup service which allows user to query for a search term online
  A list of online portals (terminology_lookup_config.json) are queried for the search term
  and the results (dict(information, iri)) are returned
  """

  def __init__(self) -> None:
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.session_timeout = 10  # Timeout in seconds for the requests send to the lookup services
    self.http_client = AsyncHttpClient(self.session_timeout)

  async def do_lookup(self,
                      search_term: str) -> list[dict[str, Any]]:
    """
    Do the lookup for the search term using the services in the terminology_lookup_config.json
    Args:
      search_term (str): Search term used for the lookup

    Returns: List of IRI information for the search term crawled online using the services in the terminology_lookup_config.json
    In case of error, an empty list is returned and session_request_errors is updated with failures

    """
    if not search_term or search_term.isspace():
      self.logger.error("Invalid null search term!")
      return []
    self.logger.info("Searching for term: %s", search_term)
    self.http_client.session_request_errors.clear()  # Clear the list of request errors before the lookup
    current_path = realpath(join(getcwd(), dirname(__file__)))
    results = list[dict[str, Any]]()
    with open(join(current_path, "terminology_lookup_config.json"), encoding="utf-8") as config_file:
      for lookup_service in load(config_file):
        lookup_service['request_params'][lookup_service['search_term_key']] = search_term
        web_result = await self.http_client.get(lookup_service['url'], lookup_service['request_params'])
        if web_result and (web_result.get('status') == 200 or web_result.get('reason') == 'OK'):
          if result := self.parse_web_result(search_term, web_result.get('result'),
                                             lookup_service):
            results.append(result)
        else:
          self.logger.error("Error while querying the lookup service: %s, Reason: %s, Status: %s, Error: %s",
                            lookup_service.get('name'),
                            web_result.get('reason'),
                            web_result.get('status'),
                            self.http_client.session_request_errors)
    return results

  def parse_web_result(self,
                       search_term: str,
                       web_result: dict[str, Any],
                       lookup_service: dict[str, Any]) -> dict[str, Any]:
    """
    Parse the web result returned by querying the lookup service for the search term
    Args:
      search_term (str): Search term used for the lookup
      web_result (dict[str, Any]): Web result returned by querying the lookup service
      lookup_service (dict[str, Any]): Lookup service taken from the terminology_lookup_config.json

    Returns (dict[str, Any]): Dictionary containing the name, search term and results which is a list of dict(iri, information)

    """
    if (not search_term or
        not web_result or
        not lookup_service):
      self.logger.error("Invalid search term or web result or lookup service!")
      return {}
    self.logger.info("Searching term: %s for online service: %s",
                     search_term,
                     lookup_service['name'])
    retrieved_results: dict[str, Any] = {
      "name": lookup_service['name'],
      "search_term": search_term,
      "results": []
    }

    # Get mandatory attributes
    result_keys = lookup_service['search_criteria']['results_keys']
    desc_keys = lookup_service['search_criteria']['description_keys']
    id_key = lookup_service['search_criteria']['id_key']

    # Get non mandatory attributes
    duplicate_ontology_names = lookup_service.get('duplicate_ontology_names')
    skip_desc = lookup_service.get('skip_description')
    duplicate_ontology_key = lookup_service['search_criteria'].get('ontology_name_key')

    results = reduce(lambda d, key: d.get(key) if d else None, result_keys,  # type: ignore[arg-type, return-value]
                     web_result)
    for item in results or []:
      description = reduce(lambda d, key: d.get(key) if d else None, desc_keys, item)  # type: ignore[attr-defined]
      is_duplicate = (item[duplicate_ontology_key]  # type: ignore[operator]
                      in duplicate_ontology_names) if duplicate_ontology_key else False
      if (description
          and description != skip_desc
          and not is_duplicate):
        retrieved_results["results"].append(
          {
            "iri": f"{lookup_service['iri_prefix']}{item[id_key]}"
            if lookup_service.get('iri_prefix') else item[id_key],
            "information": ",".join(description) if isinstance(description, list)
            else description
          }
        )
    return retrieved_results



""" Main class of config tab on authors """
import json, re
from enum import Enum
from pathlib import Path
from typing import Callable, Any
import requests
from PySide6.QtWidgets import QWidget, QFormLayout, QLabel, QLineEdit, QComboBox  # pylint: disable=no-name-in-module
from ..miscTools import restart
from ..guiStyle import TextButton, IconButton, widgetAndLayout
from ..guiCommunicate import Communicate


class ConfigurationAuthors(QWidget):
  """ Main class of config tab on authors

  FOR NOW: only one author (one self) can be added to align with GDPR
  """
  def __init__(self, comm:Communicate, callbackFinished:Callable[[bool],None]):
    """
    Initialization

    Args:
      comm (Communicate): communication
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm = comm
    self.callbackFinished = callbackFinished

    #GUI elements
    if hasattr(self.comm.backend, 'configuration'):
      self.author = self.comm.backend.configuration['authors'][0]
      self.tabAuthorL = QFormLayout(self)
      self.userOrcid = self.addRowText('orcid','ORCID')
      self.userTitle = self.addRowText('title','Title')
      self.userFirst = self.addRowText('first','First name')
      self.userLast  = self.addRowText('last', 'Surname')
      self.userEmail = self.addRowText('email','Email address')
      #headline of organizations
      orgaW, orgaL = widgetAndLayout('H', None, spacing='s', top='l')
      self.orgaCB = QComboBox()
      self.orgaCB.addItems([i['organization'] for i in self.author['organizations']])
      orgaL.addStretch(1)
      orgaL.addWidget(self.orgaCB, stretch=2)                       # type: ignore[call-arg]
      IconButton('fa5s.plus-circle', self, [Command.ADD], orgaL, 'Add organization')
      IconButton('fa5s.minus-circle', self, [Command.DELETE], orgaL, 'Delete organization')
      self.tabAuthorL.addRow(orgaW)

      self.userRorid = self.addRowText('rorid','RORID')
      self.userOrg   = self.addRowText('organization','Organization')
      buttonBarW, buttonBarL = widgetAndLayout('H', None, top='l')
      buttonBarL.addStretch(1)
      TextButton('Save changes', self, [Command.SAVE], buttonBarL)
      self.tabAuthorL.addRow(buttonBarW)
      self.orgaCB.previousIndex = 0
      self.lockSelfAuthor = False
      self.orgaCB.currentIndexChanged.connect(lambda: self.execute([Command.CHANGE])) #connect to slot only after all painting is done


  def addRowText(self, item:str, label:str) -> QLineEdit:
    """
    Add a row with a combo-box to the form

    Args:
      item (str): property name in configuration file
      label (str): label used in form

    Returns:
      QTextEdit: text-edit widget with content
    """
    rightW = QLineEdit()
    if item in {'organization','rorid'}:
      rightW.setText(self.author['organizations'][0][item] if self.author['organizations'] else '')
    else:
      rightW.setText(self.author[item])
    rightW.setAccessibleName(item)
    if item in {'organization', 'rorid','orcid'}:
      rightW.editingFinished.connect(self.changedID)
    self.tabAuthorL.addRow(QLabel(label), rightW)
    return rightW


  def changedID(self) -> None:
    """
    Autofill based on orcid and rorid
    """
    sender = self.sender().accessibleName()
    if sender == 'rorid':
      if re.match(r'^\w{9}$', self.userRorid.text().strip() ) is not None:
        reply = requests.get(
            f'https://api.ror.org/organizations/{self.userRorid.text().strip()}')
        self.userOrg.setText(reply.json()['name'])
        self.orgaCB.setItemText(self.orgaCB.currentIndex(), reply.json()['name'])
    elif sender == 'orcid':
      if re.match(r'^\w{4}-\w{4}-\w{4}-\w{4}$', self.userOrcid.text().strip() ) is not None:
        reply = requests.get(
            f'https://pub.orcid.org/v3.0/{self.userOrcid.text().strip()}')
        text = reply.content.decode()
        first = text.split('<personal-details:given-names>')[1].split('</personal-details:given-names>')[0]
        last = text.split('<personal-details:family-name>')[1].split('</personal-details:family-name>')[0]
        self.userFirst.setText(first)
        self.userLast.setText(last)
    elif sender == 'organization':
      self.orgaCB.setItemText(self.orgaCB.currentIndex(), self.userOrg.text())
    else:
      print('**ERROR: did not understand sender:',sender)
    return


  def execute(self, command:list[Any]) -> None:
    """
    Save changes to hard-disk
    """
    if command[0] is Command.SAVE:
      self.author['first'] = self.userFirst.text().strip()
      self.author['last']  = self.userLast.text().strip()
      self.author['title'] = self.userTitle.text().strip()
      self.author['email'] = self.userEmail.text().strip()
      self.author['orcid'] = self.userOrcid.text().strip()
      j = self.orgaCB.currentIndex()
      if j>-1:
        self.author['organizations'][j]['organization'] = self.userOrg.text().strip()
        self.author['organizations'][j]['rorid']        = self.userRorid.text().strip()
      self.comm.backend.configuration['authors'][0] = self.author
      with open(Path.home()/'.pastaELN.json', 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.comm.backend.configuration,indent=2))
      self.callbackFinished(False)
    elif command[0] is Command.ADD:
      self.orgaCB.addItem('- new -')
      if len(self.author['organizations'])<self.orgaCB.count():
        self.author['organizations'].append({'rorid':'', 'organization':''})
      self.orgaCB.setCurrentText('- new -')
      self.userRorid.setEnabled(True)
      self.userRorid.setText('')
      self.userOrg.setEnabled(True)
      self.userOrg.setText('')
    elif command[0] is Command.DELETE:
      k = self.orgaCB.currentIndex()
      if k>-1:
        self.author['organizations'].pop(k)
        self.lockSelfAuthor = True
        self.orgaCB.removeItem(k)
      if self.orgaCB.currentIndex()==-1:
        self.userRorid.setEnabled(False)
        self.userOrg.setEnabled(False)
      self.lockSelfAuthor = False
    elif command[0] is Command.CHANGE and self.author['organizations']:
      j = self.orgaCB.previousIndex
      if j<len(self.author['organizations']) and not self.lockSelfAuthor:
        self.author['organizations'][j]['organization'] = self.userOrg.text().strip()
        self.author['organizations'][j]['rorid']        = self.userRorid.text().strip()
      k = self.orgaCB.currentIndex()
      self.userRorid.setText(self.author['organizations'][k]['rorid'])
      self.userOrg.setText(self.author['organizations'][k]['organization'])
    self.orgaCB.previousIndex = self.orgaCB.currentIndex()
    return


class Command(Enum):
  """ Commands used in this file """
  SAVE   = 1
  ADD    = 2
  DELETE = 3
  CHANGE = 4



""" OLD Main function when command-line commands used

Called by user or react-electron frontend. Keep it simple: only functions that
are required by frontend. Otherwise, make only temporary changes
"""
import json, argparse, traceback
from pathlib import Path
from subprocess import run, PIPE, STDOUT
import urllib.request

from pasta_eln import __version__
from .backend import Backend
from .miscTools import upOut, upIn, updateExtractorList, DummyProgressBar
from .inputOutput import importELN, exportELN
from .installationTools import configuration as checkConfiguration
from .installationTools import exampleData

def commands(getDocu:bool, args:argparse.Namespace) -> str:
  """
  Main function

  Args:
    getDocu (bool): True=return documentation string; False=use arguments
    args (argparse.Namespace): arguments supplied by user / frontend

  Returns:
    string: documentation or empty string
  """
  doc = ''
  success = ''  #success ''=undecided; '-1'=False; '1'=True
  pathConfig = Path.home()/'.pastaELN.json'

  ##################################################
  ## general things that do not require open database / change configuration file
  if getDocu:
    doc += '  help: help information\n'
    doc += '    example: pastaELN_CLI.py help\n'
  elif args.command=='help':
    print("HELP:")
    return '1'

  doc += '\n-- Configuration file commands --\n'
  if getDocu:
    doc += '  verifyConfiguration: verify configuration file\n'
    doc += '    example: pastaELN_CLI.py verifyConfiguration\n'
    doc += '    example: pastaELN_CLI.py verifyConfigurationRepair (repair function)\n'
  elif args.command.startswith('verifyConfiguration'):
    command = 'repair' if args.command=='verifyConfigurationRepair' else 'test'
    output = checkConfiguration(command=command)
    print(output)
    return '-1' if '**ERROR' in output else '1'

  if getDocu:
    doc += '  extractorScan: get list of all extractors and save into .pastaELN.json\n'
    doc += '    example: pastaELN_CLI.py extractorScan\n'
  elif args.command=='extractorScan':
    with open(pathConfig,'r', encoding='utf-8') as f:
      configuration = json.load(f)
      pathToExtractors = Path(__file__).parent / 'extractors' \
        if 'extractorDir' not in configuration \
        else configuration['extractorDir']
    updateExtractorList(pathToExtractors)
    return '1'

  if not getDocu and args.command=='up':
    print('up:',upOut(args.docID))
    return '1'

  if getDocu:
    doc += '  scramble: scramble the password and user name in configuration file\n'
    doc += '    example: pastaELN_CLI.py scramble\n'
  elif args.command=='scramble':
    with open(pathConfig,'r', encoding='utf-8') as f:
      configuration = json.load(f)
    configBackup  = dict(configuration)
    links = configuration['projectGroups']
    changed = False
    for link,site in [(i,j) for i in links.keys() for j in ['local','remote']]:
      if 'user' in links[link][site] and 'password' in links[link][site]:
        links[link][site]['cred'] = upIn(links[link][site]['user']+':'+links[link][site]['password'])
        del links[link][site]['user']
        del links[link][site]['password']
        changed = True
    if changed:
      with open(Path.home()/'.pastaELN_BAK.json','w', encoding='utf-8') as f:
        f.write(json.dumps(configBackup,indent=2))
      with open(pathConfig,'w', encoding='utf-8') as f:
        f.write(json.dumps(configuration,indent=2))
    return '1'

  if getDocu:
    doc += '  decipher: decipher encrypted string\n'
    doc += '    content: string\n'
  elif args.command=='decipher':
    from serverActions import passwordDecrypt
    print(passwordDecrypt(args.content).decode())
    return '1'


  ##################################################
  ## Commands that require open PASTA database, but not specific project
  doc += '\n-- Commands that interact with backend --\n'
  try:
    # use configuration file
    if not getDocu:
      with open(pathConfig,'r', encoding='utf-8') as f:
        config = json.load(f)
      if args.database=='':
        args.database = config['defaultProjectGroup']

    initViews, initConfig, resetDataHierarchy = False, True, False
    if getDocu:
      doc += '  test: test PASTA setup\n'
      doc += '    example: pastaELN_CLI.py test -d instruments\n'
    elif args.command.startswith('test'):
      #PART 1 of test: precede with configuration test
      initViews, initConfig = True, True
      if args.command=='testDev':
        resetDataHierarchy = True
      output = "" #checkConfiguration(repair=False)  #verify configuration file .pastaELN_CLI.py
      print(output)
      if 'ERROR' in output:
        return ''
      # local and remote server test
      urls = ['http://127.0.0.1:5984']
      if config['projectGroups'][args.database]['remote']!={}:
        urls.append(config['projectGroups'][args.database]['remote']['url'])
      for url in urls:
        try:
          with urllib.request.urlopen(url) as package:
            contents = package.read()
            if json.loads(contents)['couchdb'] == 'Welcome':
              print('CouchDB server',url,'is working: username and password test upcoming')
        except Exception:
          print('**ERROR pma01: CouchDB server not working |',url)
          if url=='http://127.0.0.1:5984':
            raise NameError('**ERROR pma01a: Wrong local server.') from None

    #open backend
    if not getDocu:
      try:
        be = Backend(defaultProjectGroup=args.database, initViews=initViews, initConfig=initConfig,
                  resetDataHierarchy=resetDataHierarchy)
      except Exception:
        print('**ERROR pma20: backend could not be started.\n'+traceback.format_exc()+'\n\n')
        return ''

    if not getDocu and args.command.startswith('test') and be:
      #PART 2 of test: main test
      print('database server:',be.db.db.client.server_url)
      print('default project group:',be.configuration['defaultProjectGroup'])
      print('database name:',be.db.db.database_name)
      designDocuments = be.db.db.design_documents()
      print('Design documents')
      for item in designDocuments:
        numViews = len(item['doc']['views']) if 'views' in item['doc'] else 0
        print('  ',item['id'], '   Num. of views:', numViews )
      try:
        data = be.db.getDoc('-dataHierarchy-')
        print('DataHierarchy exists on server')
      except Exception:
        print('**ERROR pma02: DataHierarchy does NOT exist on server')
      print('local directory:',be.basePath)
      print(f'software version: {__version__}')
      return '1'

    if getDocu:
      doc += '  verifyDB: test PASTA database\n'
      doc += '    example: pastaELN_CLI.py verifyDB\n'
      doc += '    example: pastaELN_CLI.py verifyDBdev (repair function)\n'
    elif args.command.startswith('verifyDB'):
      repair = args.command=='verifyDBdev'
      output = be.checkDB(outputStyle='', repair=repair)
      print(output)
      return '1'

    if getDocu:
      doc += '  exampleData: create example data by DELETING ALL DATA\n'
      doc += '    !! BE CERTAIN THAT YOU WANT TO DO THIS !!\n'
      doc += '    example: pastaELN_CLI.py exampleData\n'
    elif args.command.startswith('exampleData'):
      #prints directly to screen
      exampleData(True)
      return '1'

    if getDocu:
      doc += '  syncLR / syncRL: synchronize with / from remote server\n'
      doc += '    example: pastaELN_CLI.py syncLR\n'
    elif args.command=='syncLR':
      progressBar = DummyProgressBar()
      success = be.replicateDB(progressBar)
      return '1' if success else '-1'
    elif args.command=='syncRL':
      be.exit()
      print('**ERROR pma03: syncRL not implemented yet')
      return '-1'

    if getDocu:
      doc += '  print: print overview\n'
      doc += "    label: possible docLabels 'Projects', 'Samples', 'Measurements', 'Procedures'\n"
      doc += "    example: pastaELN_CLI.py print -d instruments -l instrument\n"
    elif args.command=='print':
      print(be.output(args.label,True))
      return '1'

    # if getDocu:
    #   doc += '  saveBackup,loadBackup: save to file.zip / load from file.zip\n'
    #   doc += '    - docId is optional as it reduces the scope of the backup\n'
    #   doc += '    - database is optional as otherwise the default is used\n'
    #   doc += '    example: pastaELN_CLI.py saveBackup -d instruments\n'
    #   doc += '    example: pastaELN_CLI.py saveBackup -i x-76b0995cf655bcd487ccbdd8f9c68e1b\n'
    # elif args.command=='saveBackup':   #save to backup file.zip
    #   if args.docID!='':
    #     exportELN(be, args.docID, 'test.eln')
    #   else:
    #     be.backup('backup')

    # elif args.command=='loadBackup':   #load from backup file.zip
    #   be.backup('restore')
    #   return '1'

    if getDocu:
      doc += '  extractorTest: test extractor on individual datafile\n'
      doc += '    example: pastaELN_CLI.py extractorTest -i ~/[long-path]/Zeiss.tif\n'
    elif args.command=='extractorTest':
      be.testExtractor(args.docID)
      return '1'

    # if getDocu:
    #   doc += '  importXLS: import first sheet of excel file into database\n'
    #   doc += '    before: ensure database configuration and project exist\n'
    #   doc += '    example: pastaELN_CLI.py importXLS -d instruments -i x-123456 -c "~/path/to.xls" -l instrument\n'
    #   doc += '    -l is the document type\n'
    #   doc += '    afterwards: adopt dataHierarchy (views are automatically generated)\n'
    # elif args.command=='importXLS':
    #   import pandas as pd
    #   from commonTools import commonTools as cT  #not globally imported since confuses translation
    #   if args.docID!='':
    #     be.changeHierarchy(args.docID)
    #   #change for matwerks examples
    #   df = pd.read_excel(args.content, sheet_name=0)
    #   if df.shape[0]>40 and df.iloc[38].isnull().sum() < df.iloc[36].isnull().sum():
    #     df = pd.read_excel(args.content, sheet_name=0, skiprows=39, usecols=range(11))
    #     df=df.drop(df.index[[0,1]])
    #     df=df.drop(df.index[[-1,-2,-3]])
    #     #add metadata
    #     meta = pd.read_excel(args.content, sheet_name=0, skiprows=4, nrows=13, usecols=[0,2])
    #     for i in range(meta.shape[0]):
    #       key, value = meta.iloc[i,0], meta.iloc[i,1]
    #       df[key] = value
    #     #add more metadata
    #     meta = pd.read_excel(args.content, sheet_name=0, skiprows=19, nrows=8, usecols=[2,8])
    #     for i in range(meta.shape[0]):
    #       if meta.iloc[i,:].isnull().sum()==2:
    #         continue
    #       key, value = meta.iloc[i,0], meta.iloc[i,1]
    #       df[key] = value
    #     meta = pd.read_excel(args.content, sheet_name=0)
    #     df['batch'] = meta.iloc[1,0].split(' - ')[1].split(' ')[1]
    #     df['test']  = meta.iloc[1,0].split(' - ')[0][1:-1]
    #     df = df.rename(columns={"AMTwin label": "-name"})
    #   else:
    #     #default file
    #     df = pd.read_excel(args.content, sheet_name=0).fillna('')
    #   print(df.columns)
    #   print(df)
    #   for _, row in df.iterrows():
    #     data = dict((k.lower(), v) for k, v in row.items())
    #     be.addData(args.label, data )
    #   return '1'

    if getDocu:
      doc += '  redo: recreate thumbnail / use-extractor\n'
      doc += '    example: pastaELN_CLI.py redo -i m-1234567890abcdefghijklmnopqrstuv -c type/test/subtest\n'
    elif args.command=='redo':
      data = dict(be.db.getDoc(args.docID))
      if args.content is not None:
        data['-type'] = args.content.split('/')
      be.useExtractors(be.basePath/data['-branch'][0]['path'], data['shasum'], data)  #any path is good since the file is the same everywhere; data-changed by reference
      if len(data['-type'])>1 and len(data['image'])>1:
        be.db.updateDoc({'image':data['image'], '-type':data['-type']},args.docID)
        return '1'
      else:
        print('**ERROR pma06: error after redo-extraction')
        return '-1'

    if getDocu:
      doc += '  scanProject: scan project with docID\n'
      doc += '    example: pastaELN_CLI.py scanProject -i ....\n'
    elif args.command=='scanProject':
      progressBar = DummyProgressBar()
      be.scanProject(progressBar, args.docID)
      return '1'

    ##################################################
    ## Commands that require open database and open project
    doc += '\n-- Commands that interact with a special project --\n'
    if not getDocu and args.docID!='':
      be.changeHierarchy(args.docID)

    if getDocu:
      doc += '  hierarchy: print document hierarchy\n'
      doc += '    example: pastaELN_CLI.py hierarchy -i x-1234567890abc'
    elif args.command=='hierarchy':
      print(be.outputHierarchy(True,True))
      return '1'

  except:
    print(
        f"**ERROR pma10: exception thrown during pastaELN_CLI.py{traceback.format_exc()}"
        + "\n")
    raise

  if not getDocu and be is not None:
    be.exit()
  return doc

###################
## MAIN FUNCTION ##
def main() -> None:
  """
  Main function
  """
  usage = "python -m pasta_eln.cli <command> [-i docID] [-c content] [-l labels] [-d database]\n\nPossible commands are:\n"
  usage+= commands(True, argparse.Namespace())
  argparser = argparse.ArgumentParser(usage=usage)
  argparser.add_argument('command', help='see above...')
  argparser.add_argument('-i','--docID',   help='docID of project; a long alphanumeric code', default='')
  argparser.add_argument('-c','--content', help='content to save/store', default=None)
  argparser.add_argument('-l','--label',   help='label used for printing', default='x0')
  argparser.add_argument('-d','--database',help='name of database configuration', default='') #required for be = Pasta(args.database)
  arguments = argparser.parse_args()
  result = commands(False, arguments)
  if arguments.command=='help':
    argparser.print_help()
  if result == '':
    print('**ERROR pma08: command in pastaELN_CLI.py does not exist |',arguments.command)
  elif result == '1' and arguments.command!='up':
    print('SUCCESS')
  return

# start main function
if __name__ == '__main__':
  main()



""" Widget: setup tab inside the configuration dialog window """
import logging
from enum import Enum
from pathlib import Path
from typing import Callable, Any
from PySide6.QtWidgets import QWidget, QVBoxLayout, QTextEdit, QMessageBox, QFileDialog, QProgressBar    # pylint: disable=no-name-in-module
from ..guiStyle import TextButton, widgetAndLayout
from ..installationTools import couchdb, configuration, dataHierarchy, exampleData, createShortcut, installLinuxRoot
from ..fixedStringsJson import setupTextLinux, rootInstallLinux, exampleDataLinux
from ..miscTools import restart
from ..guiCommunicate import Communicate

class ConfigurationSetup(QWidget):
  """
  Main class
  """
  def __init__(self, comm:Communicate, callbackFinished:Callable[[],None]):
    """
    Initialization

    Args:
      comm (Communicate): communication
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm = comm

    #GUI elements
    self.mainL = QVBoxLayout()
    self.setMinimumWidth(400)
    self.setMinimumHeight(500)
    self.setLayout(self.mainL)
    self.callbackFinished = callbackFinished

    #widget 1 = screen 1
    self.screen1W, screen1L = widgetAndLayout('V', self.mainL)
    self.text1 = QTextEdit()
    self.mainText = setupTextLinux
    self.text1.setMarkdown(self.mainText)
    screen1L.addWidget(self.text1)
    self.progress1 = QProgressBar(self.screen1W)
    self.progress1.setMaximum(24)
    self.progress1.hide()
    screen1L.addWidget(self.progress1)

    _, footerL = widgetAndLayout('H', screen1L, 's')
    self.button1 = TextButton('Start analyse and repair', self, [Command.ANALYSE], footerL)
    self.button2 = TextButton('Finished',                 self, [Command.FINISHED], footerL)
    self.button2.hide()


  def callbackProgress(self, number:int) -> None:
    """
    Increase progressbar by moving to number

    Args:
      number (int): integer to move to
    """
    self.progress1.setValue(number)
    return


  def execute(self, command:list[Any]) -> None:
    """
    Main method that does all the analysis: open dialogs, ...
    """
    if command[0] is Command.ANALYSE:
      flagContinue = True
      logging.info('Linux setup analyse start')

      #Couchdb
      existsCouchDB = None
      res = couchdb('test')
      if res =='':
        self.mainText = self.mainText.replace('- CouchDB','- CouchDB is installed' )
        self.text1.setMarkdown(self.mainText)
        existsCouchDB = True
      else:
        existsCouchDB = False

      #Install couchdb
      if not existsCouchDB:
        button = QMessageBox.question(self, "Root installations", rootInstallLinux)
        if button == QMessageBox.Yes:
          dirName = QFileDialog.getExistingDirectory(self,'Create and select directory for scientific data',str(Path.home()))
          installLinuxRoot(existsCouchDB, Path(dirName))
          logging.info('Install linux root finished')
        else:
          self.mainText = self.mainText.replace('- CouchDB','- CouchDB: user chose to NOT install' )
          self.text1.setMarkdown(self.mainText)

      #Configuration
      if flagContinue:
        res = configuration('test')
        if res =='':
          self.mainText = self.mainText.replace('- Configuration of preferences','- Configuration of preferences is acceptable' )
          self.text1.setMarkdown(self.mainText)
        else:
          button = QMessageBox.question(self, "PASTA-ELN configuration", "Do you want to create/repair the configuration.")
          if button == QMessageBox.Yes:
            configuration('repair')
          else:
            self.mainText = self.mainText.replace('- Configuration of preferences','- Configuration: user chose to NOT install' )
            self.text1.setMarkdown(self.mainText)
            flagContinue = False

      #DataHierarchy
      if flagContinue:
        res = dataHierarchy('test')
        if '**ERROR' not in res:
          self.mainText = self.mainText.replace('- DataHierarchy of the datastructure','- DataHierarchy of the datastructure is acceptable\n'+res )
          self.text1.setMarkdown(self.mainText)
        else:
          dataHierarchy('install')

      #Shortcut
      if flagContinue:
        button = QMessageBox.question(self, "Create shortcut", "Do you want to create the shortcut for PASTA-ELN on desktop?")
        if button == QMessageBox.Yes:
          createShortcut()
          self.mainText = self.mainText.replace('- Shortcut creation', '- User selected to add a shortcut' )
        else:
          self.mainText = self.mainText.replace('- Shortcut creation', '- User selected to NOT add a shortcut' )
        self.text1.setMarkdown(self.mainText)

      #Example data
      if flagContinue:
        button = QMessageBox.question(self, "Example data", exampleDataLinux)
        if button == QMessageBox.Yes:
          self.progress1.show()
          exampleData(True, self.callbackProgress)
          self.mainText = self.mainText.replace('- Example data', '- Example data was added')
        else:
          self.mainText = self.mainText.replace('- Example data', '- Example data was NOT added, per user choice')
          flagContinue = False
        self.text1.setMarkdown(self.mainText)

      #at end
      self.button1.hide()
      self.button2.show()
      logging.info('Linux setup analyse end')
    elif command[0] is Command.FINISHED: # What do do when setup is finished: success or unsuccessfully
      restart()
      # self.callbackFinished()
    return


class Command(Enum):
  """ Commands used in this file """
  ANALYSE   = 1
  FINISHED  = 2



""" Represents the completed uploads UI. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: completed_uploads.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from datetime import datetime
from typing import Any

from PySide6 import QtCore, QtWidgets
from PySide6.QtWidgets import QDialog, QFrame

from pasta_eln.GUI.dataverse.completed_upload_task import Ui_CompletedUploadTaskFrame
from pasta_eln.GUI.dataverse.completed_uploads_base import Ui_CompletedUploadsForm
from pasta_eln.dataverse.database_api import DatabaseAPI
from pasta_eln.dataverse.upload_model import UploadModel


class CompletedUploads(Ui_CompletedUploadsForm):
  """
  Represents the completed uploads UI.

  Explanation:
      This class handles the completed uploads UI, including loading and clearing the UI,
      and creating the completed upload task widget.

  Args:
      None

  Returns:
      None
  """
  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the CompletedUploads class.

    Explanation:
        This method creates a new instance of the CompletedUploads class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the CompletedUploads class.
    """
    return super(CompletedUploads, cls).__new__(cls)

  def __init__(self)  -> None:
    """
    Initializes a new instance of the CompletedUploads class.

    Explanation:
        This method initializes a new instance of the CompletedUploads class.
        It sets up the logger, creates a QDialog instance, and sets the window modality.

    Args:
        None

    Returns:
        None
    """
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.instance = QDialog()
    super().setupUi(self.instance)
    self.db_api = DatabaseAPI()
    self.instance.setWindowModality(QtCore.Qt.ApplicationModal)

  def load_ui(self) -> None:
    """
    Loads the UI for completed uploads.

    Explanation:
        This method loads the UI for completed uploads
        by clearing the existing UI and adding widgets for each upload.
        It retrieves the completed uploads from the database and creates the corresponding widgets.

    Args:
        None

    """
    self.clear_ui()
    for upload in self.db_api.get_models(UploadModel):
      if isinstance(upload, UploadModel):
        widget = self.get_completed_upload_task_widget(upload)
        self.completedUploadsVerticalLayout.addWidget(widget)

  def clear_ui(self)  -> None:
    """
    Clears the UI for completed uploads.

    Explanation:
        This method clears the UI for completed uploads by removing all the widgets from the layout.

    Args:
        None

    """
    for widget_pos in reversed(range(self.completedUploadsVerticalLayout.count())):
      self.completedUploadsVerticalLayout.itemAt(widget_pos).widget().setParent(None)

  def get_completed_upload_task_widget(self, upload: UploadModel) -> QFrame:
    """
    Retrieves the completed upload task widget.

    Explanation:
        This function retrieves the completed upload task widget for the given upload.
        It creates a QFrame instance and sets up the UI for the completed upload task.

    Args:
        upload (UploadModel): The upload model representing the completed upload.

    Returns:
        QFrame: The completed upload task widget.
    """
    completedTaskFrame = QtWidgets.QFrame()
    completedTaskUi = Ui_CompletedUploadTaskFrame()
    completedTaskUi.setupUi(completedTaskFrame)
    completedTaskUi.projectNameLabel.setText(upload.project_name)

    completedTaskUi.statusLabel.setText(upload.status)
    match upload.status:
      case "In progress":
        completedTaskUi.dataverseUrlLabel.setText("Waiting..")
        completedTaskUi.finishedDateTimeLabel.setText("Waiting..")
      case "Finished":
        completedTaskUi.dataverseUrlLabel.setText(upload.dataverse_url)
        completedTaskUi.finishedDateTimeLabel.setText(datetime.fromisoformat(upload.finished_date_time).strftime(
          "%Y-%m-%d %H:%M:%S") if upload.finished_date_time else "")
      case "Failed" | "Cancelled":
        completedTaskUi.dataverseUrlLabel.setText("NA")
        completedTaskUi.finishedDateTimeLabel.setText("NA")
      case "Queued":
        completedTaskUi.dataverseUrlLabel.setText("Waiting..")
        completedTaskUi.finishedDateTimeLabel.setText("Waiting..")
      case _:
        completedTaskUi.dataverseUrlLabel.setText("Error state..")
        completedTaskUi.finishedDateTimeLabel.setText("Error state..")
    return completedTaskFrame


if __name__ == "__main__":
  import sys

  app = QtWidgets.QApplication(sys.argv)

  ui = CompletedUploads()
  ui.instance.show()
  sys.exit(app.exec())



""" Provides an interface to interact with the database for dataverse specific operations. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: database_api.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import logging
from json import load
from os import getcwd
from os.path import dirname, join, realpath
from typing import Any, Type, Union

from cloudant.document import Document

from pasta_eln.dataverse.base_database_api import BaseDatabaseAPI
from pasta_eln.dataverse.config_model import ConfigModel
from pasta_eln.dataverse.project_model import ProjectModel
from pasta_eln.dataverse.upload_model import UploadModel
from pasta_eln.dataverse.utils import log_and_create_error, set_authors, set_template_values


class DatabaseAPI:
  """
  Provides an interface to interact with the database for dataverse specific operations.

  Explanation:
      This class represents the DatabaseAPI and provides methods to interact with the database.
      It includes methods for initializing the database, retrieving models and data, and performing other database operations.
  """

  def __init__(self) -> None:
    """
    Initializes the DatabaseAPI instance.

    Explanation:
        This method initializes the DatabaseAPI instance
        by setting up the necessary attributes and creating an instance of BaseDatabaseAPI.

    Args:
        None

    Returns:
        None
    """
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.db_api = BaseDatabaseAPI()
    self.design_doc_name = '_design/viewDataverse'
    self.config_doc_id = '-dataverseConfig-'
    self.data_hierarchy_doc_id = '-dataHierarchy-'
    self.upload_model_view_name = "dvUploadView"
    self.project_model_view_name = "dvProjectsView"
    self.initialize_database()

  def create_dataverse_design_document(self) -> Document:
    """
    Creates a design document for the dataverse in the database.

    Explanation:
        This method creates a design document for the dataverse in the database
        using the provided design document name.

    Args:
        self: The DatabaseAPI instance.

    Returns:
        Document: The created design document.

    """
    self.logger.info("Creating design document: %s", self.design_doc_name)
    return self.db_api.create_document({"_id": self.design_doc_name})

  def create_upload_model_view(self) -> None:
    """
    Creates the dvUploadView as part of the design document.

    Explanation:
        This method creates the dvUploadView as part of the design document, using the provided design document name.

    Args:
        self: The DatabaseAPI instance.

    """
    self.logger.info("Creating dvUploadView as part of design document: %s", self.design_doc_name)
    self.db_api.add_view(self.design_doc_name, self.upload_model_view_name,
                         "function (doc) { if (doc.data_type === 'dataverse_upload') { emit(doc._id, doc); } }", None)

  def create_projects_view(self) -> None:
    """
    Creates the dvProjectsView as part of the design document.

    Explanation:
        This method creates the dvProjectsView as part of the design document, using the provided design document name.

    Args:
        self: The DatabaseAPI instance.

    """
    self.logger.info("Creating dvProjectsView as part of design document: %s", self.design_doc_name)
    self.db_api.add_view(self.design_doc_name,
                         self.project_model_view_name,
                         "function (doc) { "
                         "if (doc['-type'].includes('x0')) {"
                         "emit(doc._id, {"
                         "'name': doc['-name'], "
                         "'_id': doc._id, "
                         "'_rev': doc._rev, "
                         "'objective': doc.objective, "
                         "'status': doc.status, "
                         "'comment': doc.comment, "
                         "'user': doc['-user'], "
                         "'date': doc['-date']});"
                         "}}",
                         None
                         )

  def create_model_document(self, data: UploadModel | ConfigModel | ProjectModel) -> Union[
    UploadModel, ConfigModel, ProjectModel]:

    """
    Creates a model document in the database.

    Explanation:
        This method creates a model document in the database using the provided data.
        It removes the '_id' and '_rev' fields from the data dictionary before creating the document.

    Args:
        self: The DatabaseAPI instance.
        data (UploadModel | ConfigModel | ProjectModel): The data to be stored in the model document.

    Raises:
        ValueError: If the data parameter is None.

    Returns:
        Union[UploadModel, ConfigModel, ProjectModel]: The created model document.

    """
    self.logger.info("Creating model document: %s", data)
    if data is None:
      raise log_and_create_error(self.logger, ValueError, "Data cannot be None!")
    if not isinstance(data, (UploadModel, ConfigModel, ProjectModel)):
      raise log_and_create_error(self.logger, TypeError, "Data must be an UploadModel, ConfigModel, or ProjectModel!")
    data_dict = dict(data)
    if data_dict['_id'] is None:
      del data_dict['_id']
    del data_dict['_rev']
    return type(data)(**self.db_api.create_document(data_dict))

  def update_model_document(self, data: UploadModel | ConfigModel | ProjectModel) -> None:
    """
    Updates the model document with the provided data.

    Explanation:
        This method updates the model document with the provided data. It logs the update operation and performs necessary validations.

    Args:
        self: The DatabaseAPI instance.
        data (UploadModel | ConfigModel | ProjectModel): The data to update the model document with.

    Raises:
        ValueError: If the data parameter is None.
        TypeError: If the data parameter is not an instance of UploadModel, ConfigModel, or ProjectModel.

    """
    self.logger.info("Updating model document: %s", data)
    if data is None:
      raise log_and_create_error(self.logger, ValueError, "Data cannot be None!")
    if not isinstance(data, (UploadModel, ConfigModel, ProjectModel)):
      raise log_and_create_error(self.logger, TypeError, "Data must be an UploadModel, ConfigModel, or ProjectModel!")
    self.db_api.update_document(dict(data))

  def get_models(self, model_type: Type[Union[UploadModel, ConfigModel, ProjectModel]]) -> list[
    Union[UploadModel, ConfigModel, ProjectModel]]:
    """
    Retrieves models of the specified type from the database.

    Explanation:
        This method retrieves models of the specified type from the database using the appropriate view.

    Args:
        self: The DatabaseAPI instance.
        model_type (Type[Union[UploadModel, ConfigModel, ProjectModel]]): The type of models to retrieve.

    Raises:
        ValueError: If the model_type is not supported.

    Returns:
        list[Union[UploadModel, ConfigModel, ProjectModel]]: The retrieved models.
    """
    self.logger.info("Getting models of type: %s", model_type)
    match model_type():
      case UploadModel():
        return [UploadModel(**result) for result in self.db_api.get_view_results(self.design_doc_name, "dvUploadView")]
      case ProjectModel():
        return [ProjectModel(**result) for result in
                self.db_api.get_view_results(self.design_doc_name, "dvProjectsView")]
      case _:
        raise log_and_create_error(self.logger, TypeError, f"Unsupported model type {model_type}")

  def get_model(self, model_id: str, model_type: Type[UploadModel | ConfigModel | ProjectModel]) -> Union[
    UploadModel, ProjectModel, ConfigModel]:
    """
    Retrieves a model from the database.

    Explanation:
        This method retrieves a model from the database based on the given model_id and model_type.
        It performs the necessary validation and returns the corresponding model instance.

    Args:
        model_id (str): The ID of the model to retrieve.
        model_type (Type[UploadModel | ConfigModel | ProjectModel]): The type of the model to retrieve.

    Returns:
        Union[UploadModel, ProjectModel, ConfigModel, None]: The retrieved model instance.

    Raises:
        TypeError: If the model_type is not one of the supported types.
        ValueError: If the model_id is None.
    """
    self.logger.info("Getting model with id: %s, type: %s", model_id, model_type)
    if model_type not in (UploadModel, ProjectModel, ConfigModel):
      raise log_and_create_error(self.logger, TypeError, f"Unsupported model type {model_type}")
    if model_id is None:
      raise log_and_create_error(self.logger, ValueError, "model_id cannot be None")
    return model_type(**self.db_api.get_document(model_id))

  def get_data_hierarchy(self) -> dict[str, Any] | None:
    """
    Retrieves the data hierarchy from the database.

    Explanation:
        This method retrieves the data hierarchy from the database using the appropriate document ID.
        It performs the necessary validations and returns the retrieved data hierarchy.

    Args:
        self: The DatabaseAPI instance.

    Returns:
        dict[str, Any] | None: The retrieved data hierarchy, or None if the document is not found.
    """
    self.logger.info("Getting data hierarchy...")
    document = self.db_api.get_document(self.data_hierarchy_doc_id)
    if document is None:
      self.logger.warning("Data hierarchy document not found!")
      return document
    data = dict(document)
    del data['_id']
    del data['_rev']
    del data['-version']
    return data

  def initialize_database(self) -> None:
    """
    Initializes the database for the dataverse module.

    Explanation:
        This method initializes the database for the dataverse module by creating necessary documents and views if they do not exist.

    Args:
        self: The DatabaseAPI instance.

    """
    self.logger.info("Initializing database for dataverse module...")
    if self.db_api.get_document(self.design_doc_name) is None:
      self.create_dataverse_design_document()
    if self.db_api.get_view(self.design_doc_name, self.upload_model_view_name) is None:
      self.create_upload_model_view()
    if self.db_api.get_view(self.design_doc_name, self.project_model_view_name) is None:
      self.create_projects_view()
    if self.db_api.get_document(self.config_doc_id) is None:
      self.initialize_config_document()

  def initialize_config_document(self) -> None:
    """
    Initializes the config document.

    Explanation:
        This method initializes the config document by creating a ConfigModel instance with the provided attributes.
        It loads the metadata from a JSON file and calls the create_model_document method to create the document.

    Args:
        self: The DatabaseAPI instance.

    """
    self.logger.info("Initializing config document...")
    model = ConfigModel(_id=self.config_doc_id, parallel_uploads_count=3,
                        dataverse_login_info={"server_url": "", "api_token": "", "dataverse_id": ""},
                        project_upload_items={})
    current_path = realpath(join(getcwd(), dirname(__file__)))
    with open(join(current_path, "dataset-create-new-all-default-fields.json"), encoding="utf-8") as config_file:
      model.metadata = load(config_file)
    set_template_values(self.logger, model.metadata or {})
    set_authors(self.logger, model.metadata or {})
    self.create_model_document(model)



# Form implementation generated from reading ui file 'primitive_compound_controlled_frame_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_PrimitiveCompoundControlledFrameBase(object):
  def setupUi(self, PrimitiveCompoundControlledFrameBase):
    PrimitiveCompoundControlledFrameBase.setObjectName("PrimitiveCompoundControlledFrameBase")
    PrimitiveCompoundControlledFrameBase.resize(1345, 101)
    PrimitiveCompoundControlledFrameBase.setFrameShape(QtWidgets.QFrame.Shape.NoFrame)
    self.verticalLayout = QtWidgets.QVBoxLayout(PrimitiveCompoundControlledFrameBase)
    self.verticalLayout.setObjectName("verticalLayout")
    self.mainVerticalLayout = QtWidgets.QVBoxLayout()
    self.mainVerticalLayout.setObjectName("mainVerticalLayout")
    self.addPushButton = QtWidgets.QPushButton(parent=PrimitiveCompoundControlledFrameBase)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.addPushButton.sizePolicy().hasHeightForWidth())
    self.addPushButton.setSizePolicy(sizePolicy)
    self.addPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.addPushButton.setObjectName("addPushButton")
    self.mainVerticalLayout.addWidget(self.addPushButton)
    self.verticalLayout.addLayout(self.mainVerticalLayout)

    self.retranslateUi(PrimitiveCompoundControlledFrameBase)
    QtCore.QMetaObject.connectSlotsByName(PrimitiveCompoundControlledFrameBase)

  def retranslateUi(self, PrimitiveCompoundControlledFrameBase):
    _translate = QtCore.QCoreApplication.translate
    PrimitiveCompoundControlledFrameBase.setWindowTitle(_translate("PrimitiveCompoundControlledFrameBase", "Frame"))
    PrimitiveCompoundControlledFrameBase.setToolTip(_translate("PrimitiveCompoundControlledFrameBase", "Set the compound type for metadata block."))
    self.addPushButton.setToolTip(_translate("PrimitiveCompoundControlledFrameBase", "Add a new compound entry to the list."))
    self.addPushButton.setText(_translate("PrimitiveCompoundControlledFrameBase", "Add"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    PrimitiveCompoundControlledFrameBase = QtWidgets.QFrame()
    ui = Ui_PrimitiveCompoundControlledFrameBase()
    ui.setupUi(PrimitiveCompoundControlledFrameBase)
    PrimitiveCompoundControlledFrameBase.show()
    sys.exit(app.exec())



#!/usr/bin/python3
"""TEST using the FULL set of python-requirements: create 3 projects; simplified form of testTutorialComplex """
import os, shutil, logging
import warnings
import unittest
import tempfile
import urllib.request
import pytest
from pathlib import Path
from pasta_eln.backend import Backend
from pasta_eln.inputOutput import exportELN, importELN
from pasta_eln.miscTools import outputString
from pasta_eln.miscTools import DummyProgressBar
try:
  from eln_validator import checkFile
except:
  pass

class TestStringMethods(unittest.TestCase):
  """
  derived class for this test
  """
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.be = None
    self.dirName = ''

  @pytest.mark.skip(
    reason="Disabled for github since cannot create couchdb instance during actions")
  def test_main(self):
    """
    main function
    """
    outputFormat = 'print'
    # initialization: create database, destroy on filesystem and database and then create new one
    warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
    warnings.filterwarnings('ignore', message='invalid escape sequence')
    warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
    warnings.filterwarnings('ignore', category=ImportWarning)

    # setup
    localPath = Path('/home/steffen/FZJ/DataScience/Repositories/TheELNConsortium/TheELNFileFormat/examples')
    dirpath = localPath if localPath.exists() else Path(tempfile.mkdtemp())
    baseURL = 'https://github.com/TheELNConsortium/TheELNFileFormat/raw/master/examples/'
    files = [('kadi4mat/records-example.eln', 6), ('kadi4mat/collections-example.eln', 18),
             ('SampleDB/sampledb_export.eln', 13)]
    for fileI, numFiles in files:
      # remove everything else
      self.be = Backend('research', initConfig=False)
      self.dirName = self.be.basePath
      self.be.exit(deleteDB=True)
      shutil.rmtree(self.dirName)
      os.makedirs(self.dirName)
      self.be = Backend('research', initViews=True, initConfig=False)

      elnFile = dirpath/fileI
      if not elnFile.exists():
        print(f'Download file {elnFile}')
        elnFile.parent.mkdir(exist_ok=True)
        try:
          urllib.request.urlretrieve(baseURL+fileI, elnFile)
        except:
          print(f'Could not download eln file {baseURL+fileI} to {elnFile}')
          return

      # import
      print(f'\n\n==============================\nImport {fileI}\n==============================')
      status = importELN(self.be, str(elnFile))
      print(status)
      self.assertEqual(status[:7],'Success','Import unsuccessful')

      # test / verify after import
      print('Number of documents', len(self.be.db.getView('viewHierarchy/viewHierarchy')))
      outputString(outputFormat,'h2','VERIFY DATABASE INTEGRITY')
      outputString(outputFormat,'info', self.be.checkDB(outputStyle='text'))
      outputString(outputFormat,'h2','Tree of results on hard disk')
      os.system(f'tree {self.be.basePath}')
      outputString(outputFormat,'h2','DONE WITH VERIFY')

      fileCount = 0
      for _, _, files in os.walk(self.be.basePath):
        fileCount+=len(files)
      print(f'Number of files should:{numFiles} = reality:{fileCount}')
      self.assertEqual(fileCount, numFiles, f'Not {numFiles} files exist')
    return

  def tearDown(self):
    logging.info('End Import others test')
    return

if __name__ == '__main__':
  unittest.main()



""" Common utility functions """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: utils.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

from asyncio import CancelledError, IncompleteReadError, InvalidStateError, LimitOverrunError, \
  TimeoutError as AsyncTimeoutError
from functools import wraps
from json import JSONDecodeError
from typing import Any, Callable
from xml.etree.ElementTree import ParseError

from aiohttp import ClientConnectorError, InvalidURL
from requests.exceptions import ConnectionError as RequestsConnectionError, InvalidSchema, MissingSchema


def handle_dataverse_exception_async(wrapped: Callable[..., Any]) -> Callable[..., Any]:
  """
  Handle exceptions from dataverse client methods.
  Args:
    wrapped (Callable): The wrapped function where the exceptions can be raised.

  Returns:
    The wrapper function which handles the exceptions.
  """

  @wraps(wrapped)
  async def wrapper(self: Any, *args: object, **kwargs: object) -> Any:
    try:
      return await wrapped(self, *args, **kwargs)
    except (
        RequestsConnectionError, InvalidURL, MissingSchema, InvalidSchema, TypeError, FileNotFoundError, AttributeError,
        ParseError) as e:
      self.logger.error(e)
      return False, str(e)

  return wrapper


def handle_http_client_exception(wrapped: Callable[..., Any]) -> Callable[..., Any]:
  """
  Handle exceptions from http client methods.
  Args:
    wrapped (Callable): The wrapped function where the exceptions can be raised.

  Returns:
    The wrapper function which handles the exceptions.
  """

  @wraps(wrapped)
  async def wrapper(self: Any, *args: object, **kwargs: object) -> Any:
    try:
      return await wrapped(self, *args, **kwargs)
    except AsyncTimeoutError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session request timeout for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except CancelledError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session request cancelled for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except InvalidStateError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session request in invalid state for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except IncompleteReadError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session request incomplete read for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except LimitOverrunError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session request limit overrun for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except TypeError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session type error for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except ClientConnectorError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"ClientConnectorError for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except InvalidURL as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session InvalidURL for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except JSONDecodeError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session JSONDecodeError for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except AssertionError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session AssertionError for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}
    except ValueError as e:
      url = kwargs["base_url"] if 'base_url' in kwargs else args[0]
      error = f"Client session ValueError for url ({url}) with error: {e}"
      self.logger.error(error)
      self.session_request_errors.append(error)
      return {}

  return wrapper



""" Common functions in a number of widgets """
import platform, subprocess, os
from enum import Enum
from pathlib import Path
from typing import Any
from PySide6.QtWidgets import QMenu, QWidget  # pylint: disable=no-name-in-module
from PySide6.QtCore import     QPoint # pylint: disable=no-name-in-module
from ..guiStyle import Action


def initContextMenu(widget:QWidget, pos:QPoint) -> None:
  """
  Create a context menu

  Args:
    widget (QWidget): parent widget
    pos (position): Position to create context menu at
  """
  context = QMenu(widget)
  # for extractors
  extractors = widget.comm.backend.configuration['extractors']
  extension = Path(widget.doc['-branch'][0]['path']).suffix[1:]
  if extension.lower() in extractors:
    extractors = extractors[extension.lower()]
    baseDocType= widget.doc['-type'][0]
    choices= {key:value for key,value in extractors.items() \
                if key.startswith(baseDocType)}
    for key,value in choices.items():
      Action(value,                     widget, [CommandMenu.CHANGE_EXTRACTOR, key], context)
    context.addSeparator()
    Action('Save image',                widget, [CommandMenu.SAVE_IMAGE],            context)
  # Action('Open file with another application', widget.changeExtractor, context, widget, name='_openExternal_')
  Action('Open folder in file browser', widget, [CommandMenu.OPEN_FILEBROWSER],      context)
  Action('Hide',                        widget, [CommandMenu.HIDE],                  context)
  context.exec(widget.mapToGlobal(pos))
  return


def executeContextMenu(widget:QWidget, command:list[Any]) -> bool:
  """
  Execute context menu command

  Args:
    widget (QWidget): parent widget
    command (list): command

  Returns:
    bool: success
  """
  filePath = Path(widget.doc['-branch'][0]['path'])
  if command[0] is CommandMenu.OPEN_FILEBROWSER or command[0] is CommandMenu.OPEN_EXTERNAL:
    filePath = widget.comm.backend.basePath/filePath
    filePath = filePath if command[0] is CommandMenu.OPEN_EXTERNAL else filePath.parent
    if platform.system() == 'Darwin':       # macOS
      subprocess.call(('open', filePath))
    elif platform.system() == 'Windows':    # Windows
      os.startfile(filePath) # type: ignore[attr-defined]
    else:                                   # linux variants
      subprocess.call(('xdg-open', filePath))
  elif command[0] is CommandMenu.SAVE_IMAGE:
    image = widget.doc['image']
    if image.startswith('data:image/'):
      imageType = image[11:14] if image[14]==';' else image[11:15]
    else:
      imageType = 'svg'
    saveFilePath = widget.comm.backend.basePath/filePath.parent/f'{filePath.stem}_PastaExport.{imageType.lower()}'
    path = widget.doc['-branch'][0]['path']
    if not path.startswith('http'):
      path = (widget.comm.backend.basePath/path).as_posix()
    widget.comm.backend.testExtractor(path, recipe='/'.join(widget.doc['-type']), saveFig=str(saveFilePath))
  elif command[0] is CommandMenu.HIDE:
    widget.comm.backend.db.hideShow(widget.docID)
    widget.doc = widget.comm.backend.db.getDoc(widget.docID)
  elif command[0] is CommandMenu.CHANGE_EXTRACTOR:
    widget.doc['-type'] = command[1].split('/')
    widget.comm.backend.useExtractors(filePath, widget.doc['shasum'], widget.doc)  #any path is good since the file is the same everywhere; data-changed by reference
    if len(widget.doc['-type'])>1 and len(widget.doc['image'])>1:
      widget.doc = widget.comm.backend.db.updateDoc({'image':widget.doc['image'], '-type':widget.doc['-type']}, widget.doc['_id'])
    else:
      return False
  else:
    print(f'**ERROR: command not found in _contextMenu {command}')
  return True


class CommandMenu(Enum):
  """ Commands used in this file """
  CHANGE_EXTRACTOR = 2
  SAVE_IMAGE       = 3
  OPEN_FILEBROWSER = 4
  OPEN_EXTERNAL    = 5
  HIDE             = 6



""" widget that shows the table and the details of the items """
from typing import Any
from PySide6.QtCore import Slot                                # pylint: disable=no-name-in-module
from PySide6.QtWidgets import QWidget, QSplitter, QVBoxLayout  # pylint: disable=no-name-in-module
from .table import Table
from .details import Details
from ..guiCommunicate import Communicate

class DocTypes(QWidget):
  """ widget that shows the table and the details of the items """
  def __init__(self, comm:Communicate):
    super().__init__()
    comm.changeTable.connect(self.changeTable)
    comm.changeDetails.connect(self.changeDetails)
    # GUI elements
    table = Table(comm)
    self.details = Details(comm)
    self.details.resizeEvent = self.resizeWidget # type: ignore
    splitter = QSplitter()
    splitter.setHandleWidth(10)
    splitter.addWidget(table)
    splitter.addWidget(self.details)
    splitter.setContentsMargins(0,0,0,0)
    splitter.setSizes([1,1])
    mainL = QVBoxLayout()
    mainL.setSpacing(0)
    mainL.setContentsMargins(0, 0, 0, 0)
    mainL.addWidget(splitter)
    self.setLayout(mainL)


  @Slot(str, str)
  def changeTable(self, docType:str, projID:str) -> None:
    """
    What happens when user clicks to change doc-type
    -> show table

    Args:
      docType (str): document type
      projID (str): project ID for filtering
    """
    self.details.hide()
    return


  @Slot(str)
  def changeDetails(self, docID:str) -> None:
    """
    What happens when user clicks to change details
    -> show show details

    Args:
      docID (str): document ID
    """
    if docID!='':
      self.details.show()
    return


  def resizeWidget(self, _:Any) -> None:
    """ called if splitter resizes details-view  """
    self.details.resizeWidth(self.details.width())
    return



""" IriColumnDelegate module used for the table views """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: iri_column_delegate.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import logging
from typing import Union

from PySide6.QtCore import QModelIndex, QPersistentModelIndex
from PySide6.QtGui import Qt
from PySide6.QtWidgets import QLineEdit, QStyleOptionViewItem, QStyledItemDelegate, QWidget

from pasta_eln.GUI.data_hierarchy.lookup_iri_action import LookupIriAction


class IriColumnDelegate(QStyledItemDelegate):
  """
  Delegate for creating the line edit with lookup icon for the iri column in data hierarchy editor tables
  """

  def __init__(self) -> None:
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

  def createEditor(self,
                   parent: QWidget,
                   option: QStyleOptionViewItem,
                   index: Union[
                     QModelIndex, QPersistentModelIndex]) -> QWidget:
    """
    Creates the line edit with lookup icon for the iri column cell represented by index
    Args:
      parent (QWidget): Parent table view.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.

    Returns:

    """
    line_edit = QLineEdit(parent)
    lookup_term = None
    if index.isValid():
      # The first column value is taken as the default lookup term
      lookup_term = index.siblingAtColumn(0).data(Qt.UserRole)  # type: ignore[arg-type]
      lookup_term = lookup_term[1:] if (lookup_term
                                        and lookup_term.startswith('-')) else lookup_term
    line_edit.addAction(
      LookupIriAction(cell_index=index,
                      lookup_term=lookup_term),
      QLineEdit.TrailingPosition)
    line_edit.setClearButtonEnabled(True)
    return line_edit



""" Represents a generic task object. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: generic_task_object.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import itertools
import logging

from PySide6 import QtCore
from PySide6.QtCore import QObject


class GenericTaskObject(QObject):
  """
  Represents a generic task object.

  Explanation:
      This class represents a generic task object and provides methods for managing the task.
      It includes methods for cancelling the task, starting the task, and cleaning up the task.
  """
  cancel = QtCore.Signal()
  start = QtCore.Signal()
  finished = QtCore.Signal()
  id_iterator = itertools.count()

  def __init__(self) -> None:
    """
    Initializes the GenericTaskObject instance.

    Explanation:
        This method initializes the GenericTaskObject instance by setting the initial values of the cancelled, started, and cleaned attributes.
        It also connects the cancel signal to the cancel_task method and the start signal to the start_task method.

    Args:
        self: The GenericTaskObject instance.

    Returns:
        None
    """
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.id = next(self.id_iterator)
    self.cancelled = False
    self.started = False
    self.cleaned = False
    self.cancel.connect(lambda: self.cancel_task())  # pylint: disable=unnecessary-lambda
    self.start.connect(self.start_task)

  def cancel_task(self) -> None:
    """
    Cancels the task.

    Explanation:
        This method sets the canceled attribute of the GenericTaskObject instance to True.

    Args:
        self: The GenericTaskObject instance.

    """
    self.logger.info("Cancelling task, id: %s", self.id)
    self.cancelled = True

  def start_task(self) -> None:
    """
    Starts the task.

    Explanation:
        This method sets the cancelled attribute to False and the started attribute to True.

    Args:
        self: The GenericTaskObject instance.

    """
    self.logger.info("Starting task, id: %s", self.id)
    self.cancelled = False
    self.started = True

  def cleanup(self) -> None:
    """
    Cleans up the task.

    Explanation:
        This method sets the cleaned attribute of the GenericTaskObject instance to True.

    Args:
        self: The GenericTaskObject instance.

    """
    self.logger.info("Cleaning up task, id: %s", self.id)
    self.cleaned = True



"""extract data from vendor
- default jpeg image
"""
import base64, json
from io import BytesIO
import numpy as np
from PIL import Image

def use(filePath, recipe='', saveFileName=None):
  """
  Args:
    filePath (Path): full path file name
    recipe (string): supplied to guide recipes
                     recipe is / separated hierarchical elements parent->child
    saveFileName (string): if given, save the image to this file-name

  Returns:
    dict: containing image, metaVendor, metaUser, recipe
  """
  # Extractor
  image = Image.open(filePath)
  metaVendor = image.info
  imgArr = np.array(image)
  metaUser   = {'number pixel': imgArr.size,
                'dimension': imgArr.shape}

  #save to file
  imageData = Image.fromarray(imgArr)
  if saveFileName is not None:
    imageData.save(saveFileName)

  # convert PIL image to base64
  figfile = BytesIO()
  imageData.save(figfile, format="JPEG")
  imageData = base64.b64encode(figfile.getvalue()).decode()
  imageData = "data:image/jpeg;base64," + imageData

  # return everything
  return {'image':imageData, 'recipe':'measurement/image', 'metaVendor':metaVendor, 'metaUser':metaUser}

  #other datatypes could follow here if statements are used
  #...
  #final return if nothing successful
  #return {'recipe': '-', 'metaVendor':{}, 'metaUser':{}}



#!/usr/bin/python3
"""TEST using the FULL set of python-requirements: create 3 projects; simplified form of testTutorialComplex """
import os, shutil, logging,  json
import warnings
import unittest
from pathlib import Path
from zipfile import ZIP_DEFLATED
from zipfile import Path as ZPath
from zipfile import ZipFile
import pytest
from pasta_eln.backend import Backend
from pasta_eln.inputOutput import exportELN, importELN
from pasta_eln.miscTools import outputString
from pasta_eln.miscTools import DummyProgressBar

class TestStringMethods(unittest.TestCase):
  """
  derived class for this test
  """
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.be = None
    self.dirName = ''

  @pytest.mark.skip(
    reason="Disabled for github since cannot create couchdb instance during actions")
  def test_main(self):
    """
    main function
    """


    def test_A_File(fileName):
      """
      main function
      """
      # global variables worth discussion
      ROCRATE_NOTE_MANDATORY = ['version','sdPublisher']
      DATASET_MANDATORY = ['name']
      DATASET_SUGGESTED = ['author','mentions',  'dateCreated', 'dateModified', 'identifier', 'text', 'keywords']
      FILE_MANDATORY = ['name']
      FILE_SUGGESTED = ['sha256', 'encodingFormat', 'contentSize', 'description']

      # runtime global variables
      METADATA_FILE = 'ro-crate-metadata.json'
      OUTPUT_INFO = False
      OUTPUT_COUNTS = True
      KNOWN_KEYS = DATASET_MANDATORY+DATASET_SUGGESTED+FILE_MANDATORY+FILE_SUGGESTED+['@id', '@type']

      logJson = {}

      def processNode(graph, nodeID):
          """
          recursive function call to process each node

          Args:
            graph: full graph
            nodeID: id of node in graph
          """
          globalSuccess = True
          nodes = [ i for i in graph if '@id' in i and i['@id'] == nodeID]
          if len(nodes)!=1:
              print('**ERROR: all entries must only occur once in crate. check:', nodeID)
              return
          node = nodes[0]
          # CHECK IF MANDATORY AND SUGGESTED KEYWORDS ARE PRESENT
          if '@type' not in node:
              print('**ERROR: all nodes must have @type. check:', nodeID)
          if node['@type'] == 'Dataset':
              for key in DATASET_MANDATORY:
                  if not key in node:
                      print(f'**ERROR in dataset: "{key}" not in @id={node["@id"]}')
                      globalSuccess = False
              for key in DATASET_SUGGESTED:
                  if not key in node and OUTPUT_INFO:
                      print(f'**INFO for dataset: "{key}" not in @id={node["@id"]}')
          elif node['@type'] == 'File':
              for key in FILE_MANDATORY:
                  if not key in node:
                      print(f'**ERROR in file: "{key}" not in @id={node["@id"]}')
                      globalSuccess = False
              for key in FILE_SUGGESTED:
                  if not key in node and OUTPUT_INFO:
                      print(f'**INFO for file: "{key}" not in @id={node["@id"]}')
          # CHECK PROPERTIES FOR ALL KEYS
          if any([str(i).strip()=='' for i in node.values()]):
              print(f'**WARNING: {nodeID} contains empty values in the key-value pairs', node)
          # SPECIFIC CHECKS ON CERTAIN KEYS
          if isinstance(node.get('keywords', ''), list):
              print(f'**ERROR: {nodeID} contains an array of keywords. Use comma or space separated string')
              globalSuccess = False
          # recurse to children
          children = node.pop('hasPart') if 'hasPart' in node else []
          for child in children:
              globalSuccess = processNode(graph, child['@id']) and globalSuccess
          return globalSuccess

      print(f'\n\nParse: {fileName}')
      with ZipFile(fileName, 'r', compression=ZIP_DEFLATED) as elnFile:
          success = True
          p = ZPath(elnFile)
          dirName = sorted(p.iterdir())[0]
          metadataJsonFile = dirName.joinpath(METADATA_FILE)
          metadataContent = json.loads(metadataJsonFile.read_bytes())
          graph = metadataContent["@graph"]
          # find information from master node
          ro_crate_nodes = [i for i in graph if i["@id"] == METADATA_FILE]
          if len(ro_crate_nodes) == 1:
              for key in ROCRATE_NOTE_MANDATORY:
                  if not key in ro_crate_nodes[0]:
                      print(f'**ERROR: "{key}" not in @id={METADATA_FILE}')
          else:
              print(f'**ERROR: @id={METADATA_FILE} does not uniquely exist ')
              success = False
          main_node = [i for i in graph if i["@id"] == "./"][0]

          # iteratively go through graph
          for partI in main_node['hasPart']:
              success = processNode(graph, partI['@id']) and success
          if fileName not in logJson:
              logJson[fileName] = {'params_metadata_json':success}
          else:
              logJson[fileName] = logJson[fileName] | {'params_metadata_json':success}

          # count occurrences of all keys
          counts = {}
          for node in graph:
              if node['@id'] in ['./',METADATA_FILE]:
                  continue
              for key in node.keys():
                  if key in counts:
                      counts[key] += 1
                  else:
                      counts[key] = 1

          view = [ (v,k) for k,v in counts.items() ]
          view.sort(reverse=True)
          if OUTPUT_COUNTS:
              print('===== Counts (* unspecified)')
              for v,k in view:
                  prefix = '   ' if k in KNOWN_KEYS else ' * '
                  print(f'{prefix}{k:15}: {v}')
      print('\n\nSuccess:', success)

    # main test function: create stuff, test, ...
    outputFormat = 'print'
    # initialization: create database, destroy on filesystem and database and then create new one
    warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
    warnings.filterwarnings('ignore', message='invalid escape sequence')
    warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
    warnings.filterwarnings('ignore', category=ImportWarning)
    projectGroup = 'research'
    self.be = Backend(projectGroup, initConfig=False)

    # change documents
    docIDs = [i.split('|')[-1].strip() for i in self.be.output('procedure',True).split('\n')[2:-1]]
    docIDs +=[i.split('|')[-1].strip() for i in self.be.output('sample',True).split('\n')[2:-1]]
    for idx,docID in enumerate(docIDs):
      doc = self.be.db.getDoc(docID)
      doc['comment'] = f'Test string {idx}'
      self.be.editData(doc)

    # export
    viewProj = self.be.db.getView('viewDocType/x0')
    idProj  = [i['id'] for i in viewProj if i['value'][0]=='Intermetals at interfaces'][0]
    self.fileName = str(Path.home()/'temporary_pastaTest.eln')
    status = exportELN(self.be, idProj, self.fileName, ['procedure','measurement','sample'])
    print(f'Export to: {self.fileName}\n{status}')
    self.assertEqual(status[:21],'Success: exported 20 ','Export unsuccessful')

    # verify eln
    print('\n\nEnd export\n----------------------\nStart verification')
    test_A_File(self.fileName)

    # remove old
    docProj = self.be.db.getDoc(idProj)
    oldPath = self.be.basePath/docProj['-branch'][0]['path']
    shutil.rmtree(oldPath)
    allDocs = self.be.db.getView('viewHierarchy/viewHierarchy',    startKey=idProj)
    for doc in allDocs:
      self.be.db.remove(doc['id'])

    # import
    print('\n\n---------------\nImport')
    status = importELN(self.be, self.fileName)
    print(status)
    self.assertEqual(status[:7],'Success','Import unsuccessful')

    #test / verify
    print('Number of documents', len(self.be.db.getView('viewHierarchy/viewHierarchy',    startKey=idProj)))
    outputString(outputFormat,'h2','VERIFY DATABASE INTEGRITY')
    outputString(outputFormat,'info', self.be.checkDB(outputStyle='text'))
    outputString(outputFormat,'h2','DONE WITH VERIFY')
    fileCount = 0
    for _, _, files in os.walk(self.be.basePath):
      fileCount+=len(files)
    print(f'Number of files 15 (should be) = {fileCount} (are)')
    self.assertEqual(fileCount, 15, 'Imported entries are not 15, as they should.')
    return

  def tearDown(self):
    logging.info('End Export-import test')
    # Path(self.fileName).unlink()  #remove file
    return


if __name__ == '__main__':
  unittest.main()



""" Manages the upload queue. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: upload_manager_task_thread.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from time import sleep

from pasta_eln.dataverse.config_model import ConfigModel
from pasta_eln.dataverse.database_api import DatabaseAPI
from pasta_eln.dataverse.generic_task_object import GenericTaskObject
from pasta_eln.dataverse.task_thread_extension import TaskThreadExtension


class UploadQueueManager(GenericTaskObject):
  """
  Manages the upload queue.

  Explanation:
      This class handles the management of an upload queue.
      It provides methods to add tasks to the queue, remove tasks from the queue,
      start the task execution, clean up resources, and cancel the task.

  """

  def __init__(self) -> None:
    """
    Initializes the UploadQueueManager instance.

    Explanation:
        This method initializes the UploadQueueManager instance by setting up the logger,
        attributes for concurrent uploads,
        the upload and running queues, and the database API.
        It also calls the set_concurrent_uploads method to retrieve the number of concurrent uploads from the database.

    Args:
        None

    Returns:
        None
    """
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.number_of_concurrent_uploads: int | None = None
    self.config_model: ConfigModel | None = None
    self.upload_queue: list[TaskThreadExtension] = []
    self.running_queue: list[TaskThreadExtension] = []
    self.db_api: DatabaseAPI = DatabaseAPI()
    self.set_concurrent_uploads()

  def set_concurrent_uploads(self) -> None:
    """
    Resets the number of concurrent uploads.

    Explanation:
        This method retrieves the number of concurrent uploads from the database and sets it as the value for the
        number_of_concurrent_uploads attribute.

    Args:
        None

    """
    self.logger.info("Resetting number of concurrent uploads..")
    model = self.db_api.get_model(self.db_api.config_doc_id, ConfigModel)
    self.config_model = model if isinstance(model, ConfigModel) else None
    self.number_of_concurrent_uploads = self.config_model.parallel_uploads_count if self.config_model else None

  def add_to_queue(self, upload_task_thread: TaskThreadExtension) -> None:
    """
    Adds a thread task to the upload queue.

    Explanation:
        This method adds the given upload_task_thread to the upload queue.
        It also connects the finished signal of the task to the remove_from_queue method.

    Args:
        upload_task_thread (TaskThreadExtension): The thread task to be added to the upload queue.

    """
    self.logger.info("Adding thread task to upload queue, id: %s", upload_task_thread.task.id)
    self.upload_queue.append(upload_task_thread)
    upload_task_thread.task.finished.connect(lambda: self.remove_from_queue(upload_task_thread))

  def remove_from_queue(self, upload_task_thread: TaskThreadExtension) -> None:
    """
    Removes a thread task from the upload queue.

    Explanation:
        This method removes the given upload_task_thread from the upload queue.
        It also removes the task from the running queue and quits the worker thread.

    Args:
        upload_task_thread (TaskThreadExtension): The thread task to be removed from the upload queue.

    """
    self.logger.info("Removing thread task from upload queue, id: %s", upload_task_thread.task.id)
    if upload_task_thread in self.upload_queue:
      self.upload_queue.remove(upload_task_thread)
    if upload_task_thread in self.running_queue:
      self.running_queue.remove(upload_task_thread)
    upload_task_thread.worker_thread.quit()

  def start_task(self) -> None:
    """
    Starts the upload queue task.

    Explanation:
        This method starts the upload queue task by calling the super-class's start_task method.
        It iterates over the upload queue
        and starts the tasks if the running queue is not full and the task is not canceled.

    Args:
        None

    """
    self.logger.info("Starting upload queue..")
    super().start_task()
    while not self.cancelled:
      if self.number_of_concurrent_uploads and len(self.running_queue) < self.number_of_concurrent_uploads:
        for upload_task_thread in self.upload_queue:
          if self.cancelled or len(self.running_queue) >= self.number_of_concurrent_uploads:
            break
          if not upload_task_thread.task.started and not upload_task_thread.task.cancelled:
            self.running_queue.append(upload_task_thread)
            upload_task_thread.task.start.emit()
      sleep(0.5)

  def cleanup(self) -> None:
    """
    Cleans up the upload manager.

    Explanation:
        This method performs the cleanup of the upload manager by calling the super-class's cleanup method
        and emptying the upload queue.

    Args:
        None

    """
    self.logger.info("Cleaning up upload manager..")
    super().cleanup()
    self.empty_upload_queue()

  def empty_upload_queue(self) -> None:
    """
    Empties the upload queue.

    Explanation:
        This method empties the upload queue by quitting each upload task thread and clearing the upload queue.

    Args:
        None

    """
    self.logger.info("Emptying upload queue..")
    for upload_task_thread in self.upload_queue:
      upload_task_thread.quit()
    self.upload_queue.clear()

  def cancel_task(self) -> None:
    """
    Cancels the upload queue manager completely.

    Explanation:
        This method cancels the upload queue by calling the super-class's cancel_task method.
        It emits the cancel signal for each task in the running queue and empties the upload queue.

    Args:
        None

    """
    self.logger.info("Cancelling upload queue..")
    super().cancel_task()
    for upload_task_thread in self.running_queue:
      upload_task_thread.task.cancel.emit()
    self.empty_upload_queue()



# Form implementation generated from reading ui file 'terminology_lookup_dialog_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_TerminologyLookupDialogBase(object):
  def setupUi(self, TerminologyLookupDialogBase):
    TerminologyLookupDialogBase.setObjectName("TerminologyLookupDialogBase")
    TerminologyLookupDialogBase.resize(900, 767)
    self.gridLayout = QtWidgets.QGridLayout(TerminologyLookupDialogBase)
    self.gridLayout.setObjectName("gridLayout")
    self.buttonBox = QtWidgets.QDialogButtonBox(parent=TerminologyLookupDialogBase)
    self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
    self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
    self.buttonBox.setObjectName("buttonBox")
    self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 1)
    self.mainVerticalLayout = QtWidgets.QVBoxLayout()
    self.mainVerticalLayout.setObjectName("mainVerticalLayout")
    self.searchBarHorizontalLayout = QtWidgets.QHBoxLayout()
    self.searchBarHorizontalLayout.setObjectName("searchBarHorizontalLayout")
    self.terminologyLineEdit = QtWidgets.QLineEdit(parent=TerminologyLookupDialogBase)
    self.terminologyLineEdit.setClearButtonEnabled(True)
    self.terminologyLineEdit.setObjectName("terminologyLineEdit")
    self.searchBarHorizontalLayout.addWidget(self.terminologyLineEdit)
    self.terminologySearchPushButton = QtWidgets.QPushButton(parent=TerminologyLookupDialogBase)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.terminologySearchPushButton.sizePolicy().hasHeightForWidth())
    self.terminologySearchPushButton.setSizePolicy(sizePolicy)
    self.terminologySearchPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.terminologySearchPushButton.setObjectName("terminologySearchPushButton")
    self.searchBarHorizontalLayout.addWidget(self.terminologySearchPushButton)
    self.mainVerticalLayout.addLayout(self.searchBarHorizontalLayout)
    self.searchProgressBar = QtWidgets.QProgressBar(parent=TerminologyLookupDialogBase)
    self.searchProgressBar.setProperty("value", 0)
    self.searchProgressBar.setObjectName("searchProgressBar")
    self.mainVerticalLayout.addWidget(self.searchProgressBar)
    self.scrollAreaHorizontalLayout = QtWidgets.QHBoxLayout()
    self.scrollAreaHorizontalLayout.setObjectName("scrollAreaHorizontalLayout")
    self.scrollArea = QtWidgets.QScrollArea(parent=TerminologyLookupDialogBase)
    self.scrollArea.setWidgetResizable(True)
    self.scrollArea.setObjectName("scrollArea")
    self.scrollAreaWidgetContents = QtWidgets.QWidget()
    self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 876, 415))
    self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
    self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
    self.gridLayout_2.setObjectName("gridLayout_2")
    self.scrollAreaContentsVerticalLayout = QtWidgets.QVBoxLayout()
    self.scrollAreaContentsVerticalLayout.setObjectName("scrollAreaContentsVerticalLayout")
    self.gridLayout_2.addLayout(self.scrollAreaContentsVerticalLayout, 0, 0, 1, 1)
    self.scrollArea.setWidget(self.scrollAreaWidgetContents)
    self.scrollAreaHorizontalLayout.addWidget(self.scrollArea)
    self.mainVerticalLayout.addLayout(self.scrollAreaHorizontalLayout)
    self.errorConsoleVerticalLayout = QtWidgets.QVBoxLayout()
    self.errorConsoleVerticalLayout.setObjectName("errorConsoleVerticalLayout")
    self.errorButtonsHorizontalLayout = QtWidgets.QHBoxLayout()
    self.errorButtonsHorizontalLayout.setObjectName("errorButtonsHorizontalLayout")
    self.errorConsolePushButton = QtWidgets.QPushButton(parent=TerminologyLookupDialogBase)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.errorConsolePushButton.sizePolicy().hasHeightForWidth())
    self.errorConsolePushButton.setSizePolicy(sizePolicy)
    self.errorConsolePushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.errorConsolePushButton.setObjectName("errorConsolePushButton")
    self.errorButtonsHorizontalLayout.addWidget(self.errorConsolePushButton)
    spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.errorButtonsHorizontalLayout.addItem(spacerItem)
    self.errorConsoleVerticalLayout.addLayout(self.errorButtonsHorizontalLayout)
    self.errorConsoleTextEdit = QtWidgets.QTextEdit(parent=TerminologyLookupDialogBase)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.errorConsoleTextEdit.sizePolicy().hasHeightForWidth())
    self.errorConsoleTextEdit.setSizePolicy(sizePolicy)
    self.errorConsoleTextEdit.setMinimumSize(QtCore.QSize(0, 0))
    self.errorConsoleTextEdit.setObjectName("errorConsoleTextEdit")
    self.errorConsoleVerticalLayout.addWidget(self.errorConsoleTextEdit)
    self.mainVerticalLayout.addLayout(self.errorConsoleVerticalLayout)
    self.gridLayout.addLayout(self.mainVerticalLayout, 0, 0, 1, 1)

    self.retranslateUi(TerminologyLookupDialogBase)
    self.buttonBox.accepted.connect(TerminologyLookupDialogBase.accept) # type: ignore
    self.buttonBox.rejected.connect(TerminologyLookupDialogBase.reject) # type: ignore
    QtCore.QMetaObject.connectSlotsByName(TerminologyLookupDialogBase)

  def retranslateUi(self, TerminologyLookupDialogBase):
    _translate = QtCore.QCoreApplication.translate
    TerminologyLookupDialogBase.setWindowTitle(_translate("TerminologyLookupDialogBase", "Terminology Lookup"))
    self.terminologyLineEdit.setPlaceholderText(_translate("TerminologyLookupDialogBase", "\"Search for Definitions in Wikis/Ontologies\""))
    self.terminologySearchPushButton.setText(_translate("TerminologyLookupDialogBase", "Search"))
    self.errorConsolePushButton.setText(_translate("TerminologyLookupDialogBase", "Show/Hide Errors"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    TerminologyLookupDialogBase = QtWidgets.QDialog()
    ui = Ui_TerminologyLookupDialogBase()
    ui.setupUi(TerminologyLookupDialogBase)
    TerminologyLookupDialogBase.show()
    sys.exit(app.exec())



""" Graphical user interface includes all widgets """
import json
import logging
import os
import sys
import webbrowser
from enum import Enum
from pathlib import Path
from typing import Any, Union

from PySide6.QtCore import Qt, Slot, QCoreApplication  # pylint: disable=no-name-in-module
from PySide6.QtGui import QIcon, QPixmap, QShortcut  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import QMainWindow, QApplication, QFileDialog, QMessageBox  # pylint: disable=no-name-in-module
from qt_material import apply_stylesheet  # of https://github.com/UN-GCPDS/qt-material

from pasta_eln import __version__
from .GUI.body import Body
from .GUI.config import Configuration
from .GUI.form import Form
from .GUI.projectGroup import ProjectGroup
from .GUI.sidebar import Sidebar
from .backend import Backend
from .fixedStringsJson import shortcuts
from .guiCommunicate import Communicate
from .guiStyle import Action, showMessage, widgetAndLayout, ScrollMessageBox
from .inputOutput import exportELN, importELN
from .GUI.data_hierarchy.data_hierarchy_editor_dialog import DataHierarchyEditorDialog
from .miscTools import updateExtractorList, restart

os.environ['QT_API'] = 'pyside6'


# Subclass QMainWindow to customize your application's main window
class MainWindow(QMainWindow):
  """ Graphical user interface includes all widgets """

  def __init__(self) -> None:
    # global setting
    super().__init__()
    venv = ' without venv' if sys.prefix == sys.base_prefix and 'CONDA_PREFIX' not in os.environ else ' in venv'
    self.setWindowTitle(f"PASTA-ELN {__version__}{venv}")
    self.setWindowState(Qt.WindowMaximized)  # type: ignore
    resourcesDir = Path(__file__).parent / 'Resources'
    self.setWindowIcon(QIcon(QPixmap(resourcesDir / 'Icons' / 'favicon64.png')))
    self.backend = Backend()
    self.comm = Communicate(self.backend)
    self.comm.formDoc.connect(self.formDoc)

    # Menubar
    menu = self.menuBar()
    projectMenu = menu.addMenu("&Project")
    Action('&Export .eln',                   self, [Command.EXPORT],         projectMenu)
    Action('&Import .eln',                   self, [Command.IMPORT],         projectMenu)
    projectMenu.addSeparator()
    Action('&Exit',                          self, [Command.EXIT],           projectMenu)

    viewMenu = menu.addMenu("&Lists")
    if hasattr(self.backend, 'db'):
      for docType, docLabel in self.comm.backend.db.dataLabels.items():
        if docType[0] == 'x' and docType[1] != '0':
          continue
        shortcut = self.comm.backend.db.dataHierarchy[docType].get('shortcut','')
        shortcut = None if shortcut=='' else f"Ctrl+{shortcut}"
        Action(docLabel,                     self, [Command.VIEW, docType],  viewMenu, shortcut=shortcut)
        if docType == 'x0':
          viewMenu.addSeparator()
      viewMenu.addSeparator()
      Action('&Tags',                        self, [Command.VIEW, '_tags_'], viewMenu, shortcut='Ctrl+T')
      Action('&Unidentified',                self, [Command.VIEW, '-'],      viewMenu, shortcut='Ctrl+U')

    systemMenu = menu.addMenu("&System")
    Action('&Project groups',                self, [Command.PROJECT_GROUP],   systemMenu)
    changeProjectGroups = systemMenu.addMenu("&Change project group")
    if hasattr(self.backend, 'configuration'):                            # not case in fresh install
      for name in self.backend.configuration['projectGroups'].keys():
        Action(name,                         self, [Command.CHANGE_PG, name], changeProjectGroups)
    Action('&Synchronize',                   self, [Command.SYNC],            systemMenu, shortcut='F5')
    Action('&Data hierarchy editor',         self, [Command.DATAHIERARCHY],        systemMenu, shortcut='F8')
    systemMenu.addSeparator()
    Action('&Test extraction from a file',   self, [Command.TEST1],           systemMenu)
    Action('Test &selected item extraction', self, [Command.TEST2],           systemMenu, shortcut='F2')
    Action('Update &Extractor list',         self, [Command.UPDATE],          systemMenu)
    systemMenu.addSeparator()
    Action('&Configuration',                 self, [Command.CONFIG],          systemMenu, shortcut='Ctrl+0')

    helpMenu = menu.addMenu("&Help")
    Action('&Website',                       self, [Command.WEBSITE],         helpMenu)
    Action('&Verify database',               self, [Command.VERIFY_DB],       helpMenu, shortcut='Ctrl+?')
    Action('Shortcuts',                      self, [Command.SHORTCUTS],       helpMenu)
    # shortcuts for advanced usage (user should not need)
    QShortcut('F9', self, lambda: self.execute([Command.RESTART]))

    # GUI elements
    mainWidget, mainLayout = widgetAndLayout('H')
    self.setCentralWidget(mainWidget)  # Set the central widget of the Window
    body = Body(self.comm)  # body with information
    self.sidebar = Sidebar(self.comm)  # sidebar with buttons
    # sidebarScroll = QScrollArea()
    # sidebarScroll.setWidget(self.sidebar)
    # if hasattr(self.comm.backend, 'configuration'):
    #   sidebarScroll.setFixedWidth(self.comm.backend.configuration['GUI']['sidebarWidth']+10)
    # mainLayout.addWidget(sidebarScroll)
    mainLayout.addWidget(self.sidebar)
    mainLayout.addWidget(body)
    self.comm.changeTable.emit('x0', '')

    #check if temporary save exist: warn user
    if (Path.home()/'.pastaELN.temp').is_file():
      ret = QMessageBox.information(self, 'Information', 'There is data from a prematurely closed form. '+
              'Do you want to use it? \n\n- If you choose yes, please reopen that form and content will' +
              'be reloaded.\n\n- If you choose no, this temporary data will be removed now.',
              QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,      # type: ignore[operator]
              QMessageBox.StandardButton.Yes)
      if ret==QMessageBox.StandardButton.No:
        (Path.home()/'.pastaELN.temp').unlink()


  @Slot(str)
  def formDoc(self, doc: dict[str, Any]) -> None:
    """
    What happens when new/edit dialog is shown

    Args:
      doc (dict): document
    """
    if '_id' in doc:
      logging.debug('gui:formdoc ' + str(doc['_id']))
    elif '_ids' in doc:
      logging.debug('gui:formdoc ' + str(doc['_ids']))
    else:
      logging.debug('gui:formdoc of type ' + str(doc['-type']))
    formWindow = Form(self.comm, doc)
    ret = formWindow.exec()
    if ret == 0:
      self.comm.stopSequentialEdit.emit()
    return


  def execute(self, command: list[Any]) -> None:
    """
    action after clicking menu item
    """
    # file menu
    if command[0] is Command.EXPORT:
      if self.comm.projectID == '':
        showMessage(self, 'Error', 'You have to open a project to export', 'Warning')
        return
      fileName = QFileDialog.getSaveFileName(self, 'Save data into .eln file', str(Path.home()), '*.eln')[0]
      if fileName != '':
        status = exportELN(self.comm.backend, self.comm.projectID, fileName)
        showMessage(self, 'Finished', status, 'Information')
    elif command[0] is Command.IMPORT:
      fileName = QFileDialog.getOpenFileName(self, 'Load data from .eln file', str(Path.home()), '*.eln')[0]
      if fileName != '':
        status = importELN(self.comm.backend, fileName)
        showMessage(self, 'Finished', status, 'Information')
        self.comm.changeSidebar.emit('redraw')
        self.comm.changeTable.emit('x0', '')
    elif command[0] is Command.EXIT:
      self.close()
    # view menu
    elif command[0] is Command.VIEW:
      self.comm.changeTable.emit(command[1], '')
    # system menu
    elif command[0] is Command.PROJECT_GROUP:
      dialog = ProjectGroup(self.comm)
      dialog.exec()
    elif command[0] is Command.CHANGE_PG:
      self.backend.configuration['defaultProjectGroup'] = command[1]
      with open(Path.home() / '.pastaELN.json', 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.backend.configuration, indent=2))
      restart()
    elif command[0] is Command.SYNC:
      report = self.comm.backend.replicateDB(progressBar=self.sidebar.progress)
      showMessage(self, 'Report of syncronization', report, style='QLabel {min-width: 450px}')
    elif command[0] is Command.DATAHIERARCHY:
      dataHierarchyForm = DataHierarchyEditorDialog(self.comm.backend.db)
      dataHierarchyForm.instance.exec()
      restart()
    elif command[0] is Command.TEST1:
      fileName = QFileDialog.getOpenFileName(self, 'Open file for extractor test', str(Path.home()), '*.*')[0]
      report = self.comm.backend.testExtractor(fileName, outputStyle='html')
      showMessage(self, 'Report of extractor test', report)
    elif command[0] is Command.TEST2:
      self.comm.testExtractor.emit()
    elif command[0] is Command.UPDATE:
      reportDict = updateExtractorList(self.backend.extractorPath)
      messageWindow = ScrollMessageBox('Extractor list updated', reportDict,
                                       style='QScrollArea{min-width:600 px; min-height:400px}')
      messageWindow.exec()
      restart()
    elif command[0] is Command.CONFIG:
      dialog = Configuration(self.comm)
      dialog.exec()
    # remainder
    elif command[0] is Command.WEBSITE:
      webbrowser.open('https://pasta-eln.github.io/pasta-eln/')
    elif command[0] is Command.VERIFY_DB:
      report = self.comm.backend.checkDB(outputStyle='html', minimal=True)
      showMessage(self, 'Report of database verification', report, style='QLabel {min-width: 800px}')
    elif command[0] is Command.SHORTCUTS:
      showMessage(self, 'Keyboard shortcuts', shortcuts)

    elif command[0] is Command.RESTART:
      restart()
    else:
      print("**ERROR gui menu unknown:", command)
    return


def mainGUI() -> tuple[Union[QCoreApplication, None], MainWindow]:
  """
    Main method and entry point for commands

  Returns:
    MainWindow: main window
  """
  # logging has to be started first
  log_path = Path.home() / 'pastaELN.log'
  #  old versions of basicConfig do not know "encoding='utf-8'"
  logging.basicConfig(filename=log_path, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                      datefmt='%m-%d %H:%M:%S')
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start PASTA GUI')
  # remainder
  if not QApplication.instance():
    application = QApplication().instance()
  else:
    application = QApplication.instance()
  main_window = MainWindow()
  logging.getLogger().setLevel(getattr(logging, main_window.backend.configuration['GUI']['loggingLevel']))
  theme = main_window.backend.configuration['GUI']['theme']
  if theme != 'none':
    apply_stylesheet(application, theme=f'{theme}.xml')
  # qtawesome and matplot cannot coexist
  import qtawesome as qta
  if not isinstance(qta.icon('fa5s.times'), QIcon):
    logging.error('qtawesome: could not load. Likely matplotlib is included and can not coexist.')
    print('qtawesome: could not load. Likely matplotlib is included and can not coexist.')
  # end test coexistence
  logging.info('End PASTA GUI')
  return application, main_window


class Command(Enum):
  """ Commands used in this file """
  EXPORT    = 1
  IMPORT    = 2
  EXIT      = 3
  VIEW      = 4
  PROJECT_GROUP = 5
  CHANGE_PG = 6
  SYNC      = 7
  DATAHIERARCHY = 8
  TEST1     = 9
  TEST2     = 10
  UPDATE    = 11
  CONFIG    = 12
  WEBSITE   = 13
  VERIFY_DB = 14
  SHORTCUTS = 15
  RESTART   = 16


def startMain() -> None:
  """
  Main function to start GUI. Extra function is required to allow starting in module fashion
  """
  app, window = mainGUI()
  window.show()
  if app:
    app.exec()


# called by python3 -m pasta_eln.gui
if __name__ == '__main__':
  startMain()



"""extract data from vendor
- comma and semicolon separated table
"""
from io import StringIO
import pandas as pd
import matplotlib.pyplot as plt

def use(filePath, recipe='', saveFileName=None):
  """
  Args:
    filePath (string): full path file name
    recipe (string): supplied to guide recipes
                     recipe is / separated hierarchical elements parent->child
    saveFileName (string): if given, save the image to this file-name
  Returns:
    dict: containing image, metaVendor, metaUser, recipe
  """
  producer = 'comma separated'
  delimiter = ','
  lines = []
  skipRows = 0
  with open(filePath, encoding='unicode_escape') as fIn:
    for  _ in range(10):
      line = fIn.readline()[:-1]
      if line.startswith('#'):
        skipRows+=1
        continue
      lines.append(line)
    # files with some form of header: try 3 criteria
    if lines[0].count(';')>lines[0].count(' ') and lines[0].count(';')==lines[1].count(';') and \
                                                   lines[0].count(';')==lines[2].count(';'): #Separate by ; not ' '
      producer = 'semicolon separated'
      delimiter = ';'
    if lines[0].count(',')>lines[0].count(' ') and lines[0].count(',')==lines[1].count(',') and \
                                                   lines[0].count(',')==lines[2].count(','): #Separate by , not ' '
      producer = 'comma separated'
      delimiter = ','
  print('Producer ', producer)

  data = pd.read_csv(filePath, delimiter=delimiter, skiprows=skipRows-1)
  plt.plot(data.iloc[:,0], data.iloc[:,1],'o-')
  metaUser = {}
  metaVendor = {}
  links = []
  recipe = 'measurement/csv'

  #save to file
  if saveFileName is not None:
    plt.savefig(saveFileName, dpi=150, bbox_inches='tight')

  #convert axes to svg image
  figfile = StringIO()
  plt.savefig(figfile, format='svg')
  image = figfile.getvalue()

  # return everything
  return {'image':image, 'recipe':recipe, 'metaVendor':metaVendor, 'metaUser':metaUser, 'links':links}

  #other datatypes follow here
  #...

  #final return if nothing successful
  # return {}



""" Represents the upload configuration dialog. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: upload_config_dialog.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from typing import Any

from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import QObject
from PySide6.QtWidgets import QCheckBox, QDialog

from pasta_eln.GUI.dataverse.upload_config_dialog_base import Ui_UploadConfigDialog
from pasta_eln.dataverse.config_model import ConfigModel
from pasta_eln.dataverse.database_api import DatabaseAPI


class UploadConfigDialog(Ui_UploadConfigDialog, QObject):
  """
  Represents the upload configuration dialog.

  Explanation:
      This class represents the upload configuration dialog.
      It provides methods to load and save the configuration settings.

  """
  config_reloaded = QtCore.Signal()

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the UploadConfigDialog class.

    Explanation:
        This method creates a new instance of the UploadConfigDialog class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the UploadConfigDialog class.
    """
    return super(UploadConfigDialog, cls).__new__(cls)

  def __init__(self) -> None:
    """
    Initializes the UploadConfigDialog.

    Explanation:
        This method initializes the UploadConfigDialog class.
        It sets up the UI and initializes the necessary attributes.
    """
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.instance = QDialog()
    super().setupUi(self.instance)
    self.db_api = DatabaseAPI()
    self.numParallelComboBox.addItems(map(str, range(2, 6)))
    self.numParallelComboBox.setCurrentIndex(2)
    self.instance.setWindowModality(QtCore.Qt.ApplicationModal)
    self.config_model: ConfigModel | None = None
    self.data_hierarchy_types: list[str] = []
    self.buttonBox.button(QtWidgets.QDialogButtonBox.Save).clicked.connect(self.save_ui)
    (self.numParallelComboBox.currentTextChanged[str]
     .connect(lambda num: setattr(self.config_model, "parallel_uploads_count", int(num))))
    self.set_data_hierarchy_types()
    self.load_ui()

  def load_ui(self) -> None:
    """
    Loads the UI for the UploadConfigDialog.

    Explanation:
        This method loads the UI for the UploadConfigDialog.
        It retrieves the config model from the database and sets up the UI elements based on the model data.
    """
    self.logger.info("Loading data and initializing UI...")
    self.config_model = self.db_api.get_model(self.db_api.config_doc_id, ConfigModel)  # type: ignore[assignment]
    if self.config_model is None:
      self.logger.error("Failed to load config model!")
      return
    for widget_pos in reversed(range(self.projectItemsVerticalLayout.count())):
      self.projectItemsVerticalLayout.itemAt(widget_pos).widget().setParent(None)
    for data_type in self.data_hierarchy_types:
      is_set = self.config_model.project_upload_items and self.config_model.project_upload_items.get(data_type, False)
      check_box = QCheckBox(text=data_type,  # type: ignore[call-overload]
                            parent=self.instance,
                            checkState=QtCore.Qt.Checked
                            if is_set else QtCore.Qt.Unchecked)
      self.projectItemsVerticalLayout.addWidget(check_box)
      check_box.stateChanged.connect(lambda state, item_name=check_box.text().capitalize():
                                     self.check_box_changed_callback(state, item_name))
    self.numParallelComboBox.setCurrentText(str(self.config_model.parallel_uploads_count))

  def check_box_changed_callback(self,
                                 state: QtCore.Qt.CheckState,
                                 project_item_name: str) -> None:
    """
    Updates the project upload items based on the state of a checkbox.

    Args:
        self: The instance of the class.
        state (QtCore.Qt.CheckState): The state of the checkbox.
        project_item_name (str): The name of the project item.

    Explanation:
        This function updates the project upload items in the config model based on the state of a checkbox.
        If the config model and project upload items are not None, it updates the project item with the given name
        to be checked or unchecked based on the state of the checkbox.
    """
    if self.config_model and self.config_model.project_upload_items is not None:
      self.config_model.project_upload_items.update({project_item_name: state == QtCore.Qt.Checked})

  def set_data_hierarchy_types(self) -> None:
    """
    Sets the data hierarchy types.

    Explanation:
        This method sets the data hierarchy types based on the data hierarchy retrieved from the database.
    """
    self.logger.info("Setting data hierarchy types...")
    data_hierarchy = self.db_api.get_data_hierarchy()
    if data_hierarchy is None:
      return
    for data_type in data_hierarchy:
      if (data_type and not data_type.isspace()
          and data_type not in ("x0", "x1", "x2")):
        type_capitalized = data_type.capitalize()
        if type_capitalized not in self.data_hierarchy_types:
          self.data_hierarchy_types.append(type_capitalized)
    self.data_hierarchy_types.append("Unidentified")

  def save_ui(self) -> None:
    """
    Saves the UI settings.

    Explanation:
        This method saves the UI settings to the configuration model.
        It retrieves the current settings from the UI elements and updates the configuration model.
        It then emits the config_reloaded signal to notify other components of the updated configuration.
    """
    self.logger.info("Saving config model...")
    if self.config_model is None:
      self.logger.error("Failed to load config model!")
      return
    self.db_api.update_model_document(self.config_model)
    self.config_reloaded.emit()

  def show(self) -> None:
    """
    Shows the instance.

    Explanation:
        This method shows the instance by calling its show() method.

    Args:
        self: The instance of the class.
    """
    self.instance.show()


if __name__ == "__main__":
  import sys

  app = QtWidgets.QApplication(sys.argv)
  ui = UploadConfigDialog()
  ui.instance.show()
  sys.exit(app.exec())



""" Creates a new instance of the ConfigDialog class. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: config_dialog.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import logging
import webbrowser
from asyncio import get_event_loop
from typing import Any

from PySide6 import QtCore, QtWidgets
from PySide6.QtWidgets import QDialog, QMessageBox

from pasta_eln.GUI.dataverse.config_dialog_base import Ui_ConfigDialogBase
from pasta_eln.dataverse.client import DataverseClient
from pasta_eln.dataverse.config_model import ConfigModel
from pasta_eln.dataverse.database_api import DatabaseAPI
from pasta_eln.dataverse.utils import check_login_credentials, decrypt_data, encrypt_data, get_encrypt_key


class ConfigDialog(Ui_ConfigDialogBase):
  """
  Creates a new instance of the ConfigDialog class.

  Explanation:
  This class represents a configuration dialog.
  It initializes the dialog, sets up the logger,
  and handles various UI elements and actions related to the configuration.

  __new__(cls, *_: Any, **__: Any) -> Any

  Creates a new instance of the ConfigDialog class.

  Args:
      *_: Variable length argument list.
      **__: Arbitrary keyword arguments.

  Returns:
      Any: The new instance of the ConfigDialog class.

  __init__(self) -> None

  Initializes the ConfigDialog.

  Explanation:
  This method initializes the ConfigDialog class.
  It sets up the logger, creates an instance of QDialog,
  and performs various initialization tasks related to the configuration.

  Usage:
    config_dialog = ConfigDialog()
    config_dialog.show()
  """

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the ConfigDialog class.

    Explanation:
        This method creates a new instance of the ConfigDialog class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the ConfigDialog class.
    """
    return super(ConfigDialog, cls).__new__(cls)

  def __init__(self) -> None:
    """
    Initializes the ConfigDialog.

    Explanation:
        This method initializes the ConfigDialog class.

        It sets up the logger and creates an instance of QDialog.

    """
    self.encrypt_key: bytes | None = None
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.instance: QDialog = QDialog()
    super().setupUi(self.instance)
    self.db_api = DatabaseAPI()
    self.config_model: ConfigModel = self.db_api.get_model(self.db_api.config_doc_id,
                                                           ConfigModel)  # type: ignore[assignment]
    self.config_model.dataverse_login_info = self.config_model.dataverse_login_info or {}
    self.decrypt_api_token()

    # Initialize UI elements
    self.dataverseServerLineEdit.setText(self.config_model.dataverse_login_info.get("server_url", ""))
    self.apiTokenLineEdit.setText(self.config_model.dataverse_login_info.get("api_token", ""))
    self.dataverseListComboBox.setCurrentText(self.config_model.dataverse_login_info.get("dataverse_id", ""))
    self.dataverseLineEdit.setText(self.config_model.dataverse_login_info.get("dataverse_id", ""))

    # Setup slots
    self.dataverseServerLineEdit.textChanged[str].connect(self.update_dataverse_server)
    self.apiTokenLineEdit.textChanged[str].connect(self.update_api_token)
    self.dataverseLineEdit.textChanged[str].connect(self.update_dataverse_id)
    self.dataverseListComboBox.currentTextChanged.connect(self.update_dataverse_line_edit)
    self.buttonBox.button(QtWidgets.QDialogButtonBox.Save).clicked.connect(self.save_config)
    self.apiTokenVerifyPushButton.clicked.connect(self.verify_server_url_and_api_token)
    self.dataverseLoadPushButton.clicked.connect(self.load_dataverse_list)
    self.apiTokenHelpPushButton.clicked.connect(
      lambda: webbrowser.open("https://data.fz-juelich.de/guide/api/auth.html"))

    # Load dataverse list
    if self.dataverseServerLineEdit.text() and self.apiTokenLineEdit.text():
      self.dataverseLoadPushButton.click()

  def decrypt_api_token(self) -> None:
    """
    Decrypts the API token using the encryption key.

    Explanation:
        This method decrypts the API token using the encryption key.
        If the encryption key is not found, it removes the API token and dataverse ID from the configuration.

    Args:
        self: The instance of the class.
    """
    key_exists, self.encrypt_key = get_encrypt_key(self.logger)
    if key_exists:
      self.config_model.dataverse_login_info["api_token"] = decrypt_data(self.logger,  # type: ignore[index]
                                                                         self.encrypt_key,
                                                                         self.config_model.dataverse_login_info[
                                                                           # type: ignore[index]
                                                                           "api_token"])
    else:
      self.logger.warning(
        "No encryption key found. Hence if any API key exists, it will be removed and the user needs to re-enter it.")
      self.config_model.dataverse_login_info["api_token"] = None  # type: ignore[index]
      self.config_model.dataverse_login_info["dataverse_id"] = None  # type: ignore[index]
      self.save_config()

  def update_dataverse_server(self, new_value: str) -> None:
    """
    Updates the server URL in the dataverse login info.

    Explanation:
        This method updates the server URL in the dataverse login info with the provided new value.

    Args:
        new_value (str): The new server URL.
    """
    self.config_model.dataverse_login_info["server_url"] = new_value  # type: ignore[index]

  def update_api_token(self, new_value: str) -> None:
    """
    Updates the API token in the dataverse login info.

    Explanation:
        This method updates the API token in the dataverse login info with the provided new value.

    Args:
        new_value (str): The new API token.
    """
    self.config_model.dataverse_login_info["api_token"] = new_value  # type: ignore[index]

  def update_dataverse_line_edit(self) -> None:
    """
    Updates the dataverse line edit with the provided dataverse ID.

    Args:
        self: The instance of the class.
    """
    dataverse_id = self.dataverseListComboBox.currentData(QtCore.Qt.ItemDataRole.UserRole)
    self.dataverseLineEdit.setText(dataverse_id)

  def update_dataverse_id(self, new_value: str) -> None:
    """
    Updates the dataverse ID in the dataverse login info.

    Explanation:
        This method updates the dataverse ID in the dataverse login info with the provided new value.

    Args:
        new_value (str): The new dataverse ID.
    """
    self.config_model.dataverse_login_info["dataverse_id"] = new_value  # type: ignore[index]

  def save_config(self) -> None:
    """
    Saves the configuration in database.

    Explanation:
        This method saves the configuration by encrypting the API token.
    """
    self.logger.info("Saving config..")
    self.config_model.dataverse_login_info["api_token"] = encrypt_data(  # type:ignore[index]
      self.logger,
      self.encrypt_key,
      self.config_model.dataverse_login_info[
        # type: ignore[index]
        "api_token"])
    self.db_api.update_model_document(self.config_model)

  def verify_server_url_and_api_token(self) -> None:
    """
    Verifies the server URL and API token.

    Explanation:
        This method verifies the server URL and API token by checking if they are both provided.
        If they are not provided, it shows a warning message.
    """
    self.logger.info("Verifying server URL and API token..")
    server_url = self.dataverseServerLineEdit.text()
    api_token = self.apiTokenLineEdit.text()
    if not (server_url and api_token):
      QMessageBox.warning(self.instance, "Error", "Please enter both server URL and API token")
      return
    success, message = check_login_credentials(self.logger, api_token, server_url)
    if success:
      QMessageBox.information(self.instance, "Credentials Valid", message)
    else:
      QMessageBox.warning(self.instance, "Credentials Invalid", message)

  def load_dataverse_list(self) -> None:
    """
    Loads the dataverse list.

    Explanation:
        This method loads the dataverse list by making a request to the server and populating the dataverse list combo box with the retrieved dataverses.
    """
    self.logger.info("Loading dataverse list..")
    saved_id = self.config_model.dataverse_login_info.get("dataverse_id", "")  # type: ignore[union-attr]
    self.dataverseListComboBox.clear()
    server_url = self.dataverseServerLineEdit.text()
    api_token = self.apiTokenLineEdit.text()
    if not (server_url and api_token):
      QMessageBox.warning(self.instance, "Error", "Please enter both server URL and API token")
      return
    dataverse_client = DataverseClient(server_url, api_token)
    event_loop = get_event_loop()
    if dataverses := event_loop.run_until_complete(dataverse_client.get_dataverse_list()):
      if isinstance(dataverses, list) and isinstance(dataverses[0], dict):
        current_text = ""
        for dataverse in dataverses:
          dv_title = dataverse.get('title', None)
          dv_id = dataverse.get('id', None)
          if dv_title and dv_id:
            self.dataverseListComboBox.addItem(dv_title, userData=dataverse['id'])
            if saved_id == dv_id:
              current_text = dv_title
        self.dataverseListComboBox.setCurrentText(current_text)
      else:
        self.logger.error("Failed to load dataverse list, error: %s", dataverses)
        QMessageBox.warning(self.instance, "Error", "Failed to load dataverse list")

  def show(self) -> None:
    """
    Shows the instance.

    Explanation:
        This method shows the instance by calling its show method.
    """
    self.instance.show()


if __name__ == "__main__":
  import sys

  app = QtWidgets.QApplication(sys.argv)

  ui = ConfigDialog()
  ui.show()
  sys.exit(app.exec())
